/*

DESCRIPTION:
    Main program file for light timer using PIC16F18324 MPU.  For a full description
    of user operation refer to timer_4-1-180101-manual-user.pdf.

SYNTAX:  n/a

KNOWN BUGS:  The menu is now too complicated to use with a gesture response
             interface.
 
             The code is a rats nest and requires a rewrite, particularly with 
             addition of new features or Bluetooth interface.  A more capable
             controller is likely required.

COPYRIGHT:  (c) 2016-2019, Brian Cornell, all rights reserved.

DEPENDENCIES:
    Compiled in *Pro* mode - necessary for size & performance optimizations.

VERSION:  6.0

HARDWARE REVISION:  180123

MODIFICATION HISTORY:

    13 May 19, BWC, version 6 branch started: *experimental*
        + Added temperature run mode which turns output on for 'on-time' hours
          when temp is > / < a set value.  This feature leverages the temp
          sensor placed under the SSR since when off it should indicate ambient
          plus an offset induced by self heating and coefficient from unit's 
          idle thermal output.

          This scheme has problems.  When used for cooling it will immediately
          want to switch back on after a cycle because of self-heating effect so 
          allowances must be made for a cooling off period before rearming.  
          Heating applications are less problematic.
 
          The following modifications were made to support:
            -Added temperature run mode & menu option
            -Added menu option to set > / < evaluation
            -Added menu option to set temperature using gestures or ambient 
             sensed by unit (no gestures)
            -Added temperature sensing when AC output is off
            -Modified EEPROM storage & retrieval
            -Modified function outputAC to apply ADC calibration constants to 
             temp objects
            -Various modifications to timer0 & timer 4 ISR blocks - see in-line 
             comments
            -Various modifications to foreground code for temperature evaluation 
             & run mode control - see in-line comments

        +Added PPS clear for LED pin during initialization so boot diagnostics 
         can display code.  Necessary since watchdog & other trap reboots don't
         clear peripheral & port settings like cold boot does.
        +Sporadic refinements to existing code to improve readability or hard-
         coded dependencies, casts, etc.  Mostly limited to those areas modified 
         to support the temp feature.

    30 Apr 18, BWC, version 5.1:
        + Moved config pragmas to config.h
        + Added WDT support in timer run loop
        + Added tests for stack under/over-flow & WDT reset and halt boot
        + Added function blinkErrorCode to display boot error code on status
          LED instead of simple halt with LED on to assist with troubleshooting
        + Added conditional compile blocks to support on-off cycle testing 
          using _TEST_ON_OFF.  This feature performs normal boot and immediately 
          enters loop waiting for a single gesture to turn output on or off.
    20 Jan 18, BWC, version 5.0 branch started:
        + Added 'reset' mode
        + Changed menu timeout to time against last gesture instead of fixed time
        + Updated current sense limits for 20A sensor and set limit @ 15.5A
    1 Jan 18, BWC, updated to use 20A current sensor 100mV/A and to allow 15A
        continuous operation.  Changed menu / option selection timeout to reset
        each time a gesture is made so indefinite time is available to user provided
        a gesture is made prior to timeout.  See function SelectOption for details.
    19 Apr 17, BWC, updated 3.x development branch based on release 3.1.
    29 Mar 17, BWC, Released as version 3.1.
    8 Feb 17, BWC, completed version 3.0 and finalized.  Supports hwr170107.
    11 Dec 16, BWC, development version 3.0d started - port from PIC12F1612->PIC16F1764
    1 Dec 16, BWC, development version 2.2d started
    19 Nov 16, BWC, development version 2.1d started
    16 Nov 16, BWC, post 2.0 development release 2.0d (I know, bad versioning) started
    5 Nov 16, BWC, release 2.0 finalized
    6 Sep 16, BWC, release 2.0 started.
    1 May 16, BWC, release 1.4 development started
    17 Apr 16, BWC, first release incorporating flash storage of user settings
    4 Apr 16, BWC, first release using light sense method.
    4 Jan 16, BWC, development release.

*/


//#define _16BIT                          //Uncomment to use 16-bit math
#define _24BIT                          //Uncomment to use 24-bit math

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <pic.h>
#include <xc.h>
#include <binary.h>
#include "config.h"
#include "main.h"

//-----------------------DATA MEMORY
unsigned int ElapsedMinutes = 0,        //Time keeping:  number of minutes that have elapsed in a 24h cycle
        OnTime = 0;                     //Time in hours that output is on
int DarkThreshold = DFLTDKTHD,          //System-wide ADC value for light sensor representing dark light level
        RtDarkThreshold = DFLTDKTHD,    //Run-time ADC value for dark threshold used to check/set value in menu
        SensedLightValue,               //ADC value representing last light sample taken
        isrTemp = TMPINVALD,            //ADC value representing last averaged temperature sample assigned in ISR
        ADCAdjFactor,                   //ADC adjustment value returned from ADC calibration routine to be applied to measurement constants
        ACurrentLimit,                  //Limiting current value with ADCAdjFactor applied
        ACurrentOffset,                 //Value representing 0A RMS - compensated for current sense 0A RMS offset
        AcRelayLimit,                   //Current value to switch relay on with ADCAdjFactor applied
        adcSetTemp;                     //ADC equivalent of SetTemp value
unsigned char ElapsedSeconds = 0,       //Time keeping:  number of elapsed seconds in current minute
        LightState = LIGHT,             //Represents current ambient light status - LIGHT or DARK
        CurrentLightState,              //Working storage for function GetLightState
        CountLightState = 0,            //Number of times light state has changed since last reset
        CountLineState = 0,             //Number of times AC line state has changed since last reset
        RunFlags,                       //Working copy in RAM of run mode flags
        OptFlags,                       //Working copy in RAM of option mode flags
        SecondsCounter = 0,             //Utility timer; each increment represents 2S
        LEDTicks,                       //Timer for controlling LED blink rate
        ACSTicks = 0,                   //Countdown timer for enabling ACS feature
        ACSMinutes = 0,                 //Countdown timer for on time of appliance override feature
        ADacqCtrlFlag = 0,              //Signals ISR to stop using ADC so can be used elsewhere
        SetTemp = 0;                    //Temperature to turn on at

//Bit status flags
bit statACoutput,                       //TRUE if AC output SHOULD be on:  controlled by function outputAC()
        statFault,                      //True if fault
        FlagCfrmACS,                    //True if timing to enable ACS
        Flag24h,                        //Indicates when 24h have elapsed
        FlagTemp;                       //True if first sample after output turned off

/*Reserve space in EEPROM to hold user settings.
 * 
 * Mapping:
 * byte[0]=CFGCAFLAG, initialized to DFLTFLAGS
 * byte[1]=CFGCATIME, initialized to DFLTONTIME
 * integer[1]=CFGIADARK, initialized to DFLTDKTHD
 * Remainder unused.
 * 
 * The union memory structure below and in subsequent store/retrieval sections
 * of the application is carry-forward from earlier versions based on PICs
 * without EEPROM.  Since v3.0 is the first release using EEPROM, resources
 * aren't in short supply, and still some uncertainty on need for future PIC 
 * change, so this scheme is left in place.  This should be overhauled in future
 * EEPROM-based releases.
 * 
 */
__eeprom union {
    unsigned char byte[ROWSIZE];
    unsigned int integer[ROWSIZE / sizeof(int)];
} FlashBlock = {DFLTFLAGS,DFLTONTIME,DFLTDKTHDL,DFLTDKTHDH, DFLTTEMP, DFLTOPTFLG};



/* Function main.  Control program for timer application.
 * 
 * Returns:  n/a; never returns
 *
 * Parameters:  n/a; called by C root
 *
*/
void main(void)
{
    static bit flagCountGestures,       //TRUE if looking for gestures to clear fault condition
               flagCool;                //TRUE if 'cooling' down after completing cycle in temp mode
    unsigned int ElapsedTime;           //Copy of system time (minutes)
    int SensedTemp;                     //Copy of isrTemp (ADC averaged value)
    unsigned char WorkingLightState,    //Stores light state from GetLightState()
            FlagDayOrNight,             //Indicates if day or night time
            FlagCfrmNight,              //True if night has been detected and timing to confirm
            FlagCfrmDay,                //True if day has been detected and timing to confirm
            defDark = DARK,             //Supports invert:  def of DARK to evaluate WorkingLightState
            defLight = LIGHT,           //Supports invert:  def of LIGHT to evaluate WorkingLightState
            lastCount,                  //Used to initiate gesture detection in fault clearing logic
            timeCountGestures;          //Countdown timer for fault clearing gestures
    union {
        unsigned char byte[ROWSIZE];
        unsigned int integer[ROWSIZE / sizeof(int)];
    } FlashBuffer;                      //Buffer to hold configuration for writing to flash


    /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     *                          Initialize timer
     * 
     * Configure MCU peripherals, interrupts, ADC calibration, and diagnostics.
     *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    Initialize();

#ifndef _TEST_ON_OFF
    /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     * Load user configurable run-time settings from flash and set control flags.
     * 
     * Variable DarkThreshold is module-wide variable used to represent the dark
     * threshold value.  However, the menu runs with default dark threshold value 
     * and not that stored in flash to ensure the menu is always navigable.  
     * So instead load flash threshold to variable RtDarkThreshold to support dark 
     * threshold check/set options.  Restore when menu exits.
     *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    OnTime = FlashBlock.byte[CFGCATIME];
    RunFlags = FlashBlock.byte[CFGCAFLAG];
    RtDarkThreshold = FlashBlock.integer[CFGIADARK];
    SetTemp = FlashBlock.byte[CFGCATEMP];
    OptFlags = FlashBlock.byte[CFGCAOFLG];
    RunFlags &= CFGFNCHGD;              //Clear config change flag
    FlagCfrmNight = FALSE;              //Not confirming switch to NGTTIME
    FlagCfrmDay = FALSE;                //Not confirming switch to DAYTIME
#ifndef _SIMUL
    GetUserAttention();                 //Tell user menu is starting...
    MainMenu();                         //...and enter menu.
#endif
    DarkThreshold = RtDarkThreshold;    //Set threshold used during run-time

    /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     * See if new settings need to be written to NVM and setup FlashBuffer for 
     * write.
     *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    if (RunFlags & CFGFYCHGD)
    {
        FlashBuffer.byte[CFGCATIME] = (unsigned char) OnTime;
        FlashBuffer.byte[CFGCAFLAG] = RunFlags;
        FlashBuffer.integer[CFGIADARK] = (unsigned int) DarkThreshold;
        FlashBuffer.byte[CFGCATEMP] = SetTemp;
        FlashBuffer.byte[CFGCAOFLG] = OptFlags;
        eepromWriteBytes((unsigned char) &FlashBlock, (unsigned char *) &FlashBuffer, sizeof(FlashBuffer));
    }

    /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     * Now run with user-specified settings.  Variable OnTime must first be
     * converted from hours to minutes.  If in night mode with dark override 
     * enabled set OnTime to HOURSDSKDN to drive control logic in timer run loop.
     * 
     * In CFGFNMODE assume starting in DAYTIME to ensure that timer starts proper 
     * operation with night mode for both timed & HOURSDSKDN.  Side effect is 
     * that light turn-on will be delayed for NGTDELAY time while darkness is 
     * confirmed.
     * 
     * 29 Mar 17 (v3.1):  in night mode, with or without dark override, set
     * Flag24h to drive timing logic in run loop.  See detailed explanation
     * at start of timer run loop.
     * 
     * If night mode invert is enabled the interpretation of DAY & NIGHT is 
     * flipped so lights operate during daytime.  This is ignored in Timed
     * or Reset modes.
     * 
     * If in Reset mode then variable ElapsedMinutes must be updated to 
     * MININDAY - OnTime so that output cycles at appropriate time.  Also 
     * disable appliance override (not allowed in this mode; doesn't make sense).
     * 
     * Timed & reset modes turn output on immediately.
     * 
     * Assign calibrated ADC value of set temperature regardless of mode using
     * macro assignTemp().
     *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    if (!(RunFlags & (CFGFTMODE | CFGRMODEO | CFGTMODEO)))  //Night mode set
    {
        if (RunFlags & CFGFEDARK) OnTime = HOURSDSKDN;
        Flag24h = TRUE;
    }
    
    OnTime = OnTime * MININHOUR;                //Convert hours to equivalent minutes
    FlagDayOrNight = DAYTIME;
    
    if (RunFlags & CFGFEIVRT)                   //Flip definition
    {
        defDark = LIGHT;
        defLight = DARK;
    }

    //Update system time & disable appliance override if Reset mode ON
    if (RunFlags & CFGRMODEO)
    {
        di();
        ElapsedMinutes = MININDAY - OnTime;
        ei();
        RunFlags &= CFGFDAPPL;
    }

    //Output on for Timed or Reset modes
    if (RunFlags & (CFGFTMODE | CFGRMODEO)) outputAC(TURNON);

    //Assign calibrated ADC temperature from user-entered temp
    assignTemp();

    /*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     *                          Timer run loop
     * 
     * All mode & turn-on/off control happens here.  This loop continues until
     * unit is reset (e.g. power cycle).  The menu is not accessible.
     * 
     * v4.0:  initialize variables for fault clearing gesture functionality.
     * 
     * v5.1:  enable watchdog timer.  WDT is used during run loop for safety
     * in the event of software failure to prevent undesired and/or unsafe
     * operation of attached equipment.
     * 
     * WDT is reset (cleared) at the top of the run loop.
     **+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    lastCount = CountLightState;
    flagCountGestures = FALSE;
    SWDTEN = TRUE;
#else
    ResetCountLightState();
#endif

    while (TRUE)
    {
#ifdef _TEST_ON_OFF
        if (CountLightState > 1)
        {
            if (statACoutput) outputAC(TURNOFF);
            else outputAC(TURNON);
            ResetCountLightState();
        }
#else
        CLRWDT();                       //Clear watchdog timer

        /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
         * 29 Mar 17 (v3.1):  time retrieval from system chrono timer.  The 
         * local object ElapsedTime is used to store the system time for 
         * evaluation in the run loop.  When retrieving time from ElapsedMinutes 
         * interrupts must be disabled since this is a 16-bit value.
         * 
         * The 24h cycle flag, Flag24h (maintained by timer1 ISR), ensures that 
         * the value assigned to ElapsedTime never exceeds MININDAY.
         * 
         * This is done for a few reasons.  First, in night mode without dark
         * override the desired behavior is to prevent the timer from starting 
         * another run cycle until within a short time of the start of the prior
         * cycle - essentially 24h-'short time'.  This prevents false starts.
         * In the DAYTIME portion of the cycle detection of NGTTIME is prevented
         * until this time.  Setting Flag24h prior to start of run loop ensures 
         * that when the timer is first started it will immediately begin 
         * sensing NGTTIME.
         * 
         * Second, the system time keeping routine clears ElapsedMinutes when
         * 24h is reached and sets the 24h flag.  If the timer is set to night
         * mode and started (e.g. powered on) in the middle of the day it 
         * wouldn't sync with the natural light rhythm due to the inhibit 
         * behavior.
         * 
         * Third, with dark override enabled and a dark environment exceeding
         * 48h the timer would complete the cycle.  We also don't want 
         * ElapsedTime to roll over to zero when in night mode since that would 
         * improperly prevent detection of NGTTIME.  Even though it is ~ 45d it 
         * could happen in places like Alaska.  
         * 
         * IMPORTANT:  Flag24h must be cleared when starting a timing cycle - 
         * both timed & dark override.  For modes, like temperature, that 
         * don't use it must still be appropriately cleared or control logic 
         * won't work properly.
         * 
         * For v6.0 the local object SensedTemp is assigned while interrupts
         * are disabled since it's a 16-bit value.
         *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
        if (Flag24h) ElapsedTime = MININDAY;
        else
        {
            di();
            ElapsedTime = ElapsedMinutes;
            SensedTemp = isrTemp;
            ei();
        }

        /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
         * Toggle status LED to indicate what's going on:  fault, appliance 
         * override, timing cycle run, idle.  Variable LEDTicks is maintained by
         * the timer0 ISR.  No precautions are taken to set value since it is a
         * single byte and non-critical in function.
         * 
         * v4.0:  provision to clear fault state without waiting for next
         * timing cycle or power cycling unit.  A series of OPTRSTFLT light state
         * changes (or OPTRSTFLT/2 gestures) is required within TMCNTGEST time 
         * period; otherwise ignored.  If timing cycle is in progress attempt
         * is made to turn back on; otherwise clear fault flag.
         *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
        if (statFault)                              //Fault
        {
            if (LEDTicks > DIAGFAULT)
            {
                LEDTicks = 0;
                ledFLIP();
                if (flagCountGestures)
                {
                    if (!(--timeCountGestures))
                    {
                        flagCountGestures = FALSE;
                        lastCount = CountLightState;
                        if (lastCount == (OPTRSTFLT - 1))
                        {
                            if (statACoutput) outputAC(TURNON);
                            else statFault = FALSE;
                        }
                    }
                }
                else if (lastCount != CountLightState)
                {
                    flagCountGestures = TRUE;
                    timeCountGestures = TMCNTGEST;
                    ResetCountLightState();
                }
            }
        }
        else if ((statACoutput) && (IOCEACS))       //Appliance override
        {
            if (LEDTicks > DIAGACS)
            {
                LEDTicks = 0;
                ledFLIP();
            }
        }
        else if (statACoutput)                      //Timing cycle
        {
            if (LEDTicks > DIAGRUN)
            {
                LEDTicks = 0;
                ledFLIP();
            }
        }
        else                                        //Idle
        {
            if (LEDTicks < DIAGIDLE) ledON(); else ledOFF();
        }

        /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
         * Appliance override control block.  Consists of enablement delay, 
         * command sense, and turn-off sections.
         * 
         * Enablement delay occurs whenever the timer completes an ON cycle - 
         * either via appliance override or normal timing event.  The delay 
         * ensures that the feature doesn't false trigger due to a noisy turn-
         * off.
         * 
         * Command sense detects line state changes and commands a turn-on.  In 
         * order for appliance override to work, two line state changes must be 
         * sensed - e.g. on->off->on or vice versa.  In order to prevent a false 
         * trigger these events must happen in a relatively short period of time.
         * ACSTicks is used for this.
         * 
         * Turn-off is executed when the ACSMinutes timer expires.
         * 
         * In order for appliance override to interact appropriately with other
         * timer functions the following conventions must be adhered to:
         *      + Macro DisableACS() must be called by user code controlling other
         *        timing turn-on events.  Not doing so will cause the turn-off
         *        code in this section to prematurely terminate the run.
         *        Conversely, this macro should not be called as part of appliance
         *        override turn-on.
         *
         *        Last, it's acceptable for an appliance override event to merge
         *        with other timing turn-on events provided these conventions
         *        are followed; the turn-off control logic in those sections 
         *        should complete the timing run normally and allow proper restart
         *        of appliance override.
         *      + Function outputAC must be called for all turn-on & off control.
         *        It contains statements that control appliance override restart
         *        state when turn-off completes.
         *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
        if (statACoutput == FALSE)
        {
            if (RunFlags & CFGFEAPPL)   //Appliance override feature must be enabled
            {
                if (!(IOCEACS))         //If not set means timer just completed cycle
                {
                    if (FlagCfrmACS)    //Check if waiting to re-enable...
                    {
                        if (!(ACSTicks)) EnableACS();   //...when timer expires
                    }
                    else                //Activate re-enable delay
                    {
                        FlagCfrmACS = TRUE;
                        ACSTicks = ACSONDELAY;
                    }
                }
                else                    //Appliance override enabled
                {
                    /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
                     * A single line state change occurred and timer expired so 
                     * reset.  This block will also reset control logic on first
                     * entry after re-enablement above.
                     *++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
                    if ((FlagCfrmACS) && !(ACSTicks))
                    {
                        FlagCfrmACS = FALSE;
                        ResetCountLineState();
                    }

                    //More than 1 line state change occurred within time limit
                    if (CountLineState > 1)
                    {
                        outputAC(TURNON);           //So turn output on...
                        ACSMinutes = ACSONTIME;     //...and set timer
                    }
                    //But if only 1 detected start timer
                    else if ((CountLineState) && !(FlagCfrmACS))
                    {
                        FlagCfrmACS = TRUE;
                        ACSTicks = ACSSWDELAY;
                    }
                }
            }
        }
        /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
         * If IOCEACS & timer are set then in appliance override mode.  Turn 
         * output off when ACSMinutes expires.
         *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
        else if ((IOCEACS) && !(ACSMinutes)) outputAC(TURNOFF);

        /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
         *                      - Timed mode (CFGFTMODE) -
         * 
         * Timed mode does not use light sensor.  Light cycling is done on a 
         * 24-hour period timed from unit power-up.  On time used was set during 
         * setup.
         *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
        if (RunFlags & CFGFTMODE)
        {
            /*Check if time to cycle timer off; turns off 1 minute after elapsed 
             * time.  But, don't change output when appliance over-ride is on.  
             * Let appliance override control block handle.
             */
            if ((statACoutput) && (ElapsedTime > OnTime) &&\
                    !(IOCEACS)) outputAC(TURNOFF);

            //If day has elapsed start new timing cycle
            if (Flag24h)
            {
                Flag24h = FALSE;
                DisableACS();           //Disable to prevent IOCEACS code from turning off
                outputAC(TURNON);
            }
        }
        /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
         *                      - Reset mode (CFGRMODEO) -
         * 
         * Reset mode does not use light sensor.  Output is on for entire 24h
         * cycle except a short 'reset' period defined by constant RESETTIME.
         * 
         * When a 24h cycle completes the output is turned off and a new timing
         * cycle is started.  After RESETTIME seconds the output is turned on
         * again.
         *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
        else if (RunFlags & CFGRMODEO)
        {
            if (!(statACoutput) && (ElapsedTime > RESETTIME)) outputAC(TURNON);
            if (Flag24h)
            {
                Flag24h = FALSE;
                outputAC(TURNOFF);
            }
        }
        /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
         *                  - Temperature mode (CFGTMODEO) -
         * 
         * Version 6.0 addition.
         * 
         * The temperature mode doesn't rely on 24h cycles and uses the system
         * time to time the on & cool-down times.  It always clears Flag24h to 
         * ensure that time & temperature are retrieved from the system (ISR).
         * 
         * Like timed mode it ensures that appliance override isn't active
         * before turning off.  Once off it employs a cool-down time indicated
         * by flagCool to help prevent false turn-ons immediately following a 
         * cycle when the temp evaluation is set >.  Unfortunately this doesn't 
         * work after an appliance override cycle since this code block can't 
         * tell the turn-off occurred without hacking those execution blocks.
         * 
         * SensedTemp is used to evaluate turn-on (instead of statACoutput) 
         * because valid temperatures are only available when the output is off.
         * Plus, immediately after turn-off a valid temp value isn't ready
         * until sampling is complete; hence evaluating by statACoutput
         * could trigger false turn-on.
         *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
        else if (RunFlags & CFGTMODEO)
        {
            Flag24h = FALSE;
            if (statACoutput)
            {
                if ((ElapsedTime > OnTime) && !(IOCEACS))
                {
                    outputAC(TURNOFF);
                    rsysTime();             //Clear to time TEMPTIME
                    flagCool = TRUE;        //Set cool-down flag
                }
            }
            else if ((flagCool) && (ElapsedTime > TEMPTIME)) flagCool = FALSE;
            else if (SensedTemp != TMPINVALD)
            {
                if (((SensedTemp > adcSetTemp) && !(OptFlags & CFGOFTMPL)) ||
                        ((SensedTemp < adcSetTemp) && (OptFlags & CFGOFTMPL)))
                {
                    DisableACS();           //Disable to prevent IOCEACS code from turning off
                    rsysTime();             //Clear to time OnTime
                    outputAC(TURNON);
                }
            }
        }
        else
        {
            /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
             *                  - Night mode (CFGFNMODE) -
             * 
             * More complex than timed mode since it can use a combination of 
             * light level and time to control lights.
             * 
             * Light turn-on is always triggered by a sustained DARK light level.
             * Turn-off can be timed or by a sustained LIGHT light level.
             *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/

            //Always start by getting current light state
            WorkingLightState = GetLightState();

            /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
             * During DAYTIME look for onset of night.  When DARK is detected 
             * start timer and set flag FlagCfrmNight.  WorkingLightState must
             * be DARK for entire timing cycle to change to NGTTIME; any
             * disruption resets.  This minimizes the chance of false trigger
             * when ambient light conditions are near dark threshold.
             * 
             * If timer expires and still DARK set NGTTIME.
             * 
             * 24 Mar 17 (v3.1):  prevent flip to NGTTIME if CFGFEDARK (dark
             * override) is disabled and time since start of last cycle is less
             * than INHIBITNGT minutes.  This prevents false triggering of cycle
             * due to transient light conditions.
             *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
            if (FlagDayOrNight == DAYTIME)
            {
                if (!(RunFlags & CFGFEDARK) && (ElapsedTime < INHIBITNGT)) 
                    continue;

                if (!(FlagCfrmNight))               //Not waiting to confirm DARK...
                {
                    if (WorkingLightState == defDark)   //...but if DARK:
                    {
                        ResetSecondsCounter();      //Start timer...
                        FlagCfrmNight = TRUE;       //... and set flag to track
                    }
                }
                else                                //Timing to confirm DARK
                {
                    if (SecondsCounter > NGTDELAY)  //Timer has expired
                    {
                        FlagCfrmNight = FALSE;      //Reset flag regardless
                        if (WorkingLightState == defDark)
                        {
                            /*++++++++++++++++++++++++++++++++++++++++++++++++++
                             * Nighttime confirmed.  CFGFNMODE doesn't care about 
                             * keeping chronological time since timer powered-up;
                             * it only cares about the elapsed time since output
                             * turn-on.
                             * 
                             * V6.0:  reset system time using macro rsysTime().
                             *++++++++++++++++++++++++++++++++++++++++++++++++*/
                            FlagDayOrNight = NGTTIME;
                            rsysTime();
                            DisableACS();   //Disable to prevent IOCEACS code from turning off
                            outputAC(TURNON);
                        }
                    } else if (WorkingLightState == defLight) FlagCfrmNight = FALSE;
                }
            }
            else
            {
                /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
                 * Flag FlagDayOrNight set to NGTTIME.  If timing lights in 
                 * CFGFNMODE, turn them off after user programmed time has 
                 * elapsed unless appliance override is on.
                 * 
                 * If OnTime value is set to (HOURSDSKDN * MININHOUR) then 
                 * ElapsedTime can never be greater than so off time will be
                 * governed by LIGHT.
                 *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
                if ((statACoutput) && (ElapsedTime > OnTime) &&\
                        !(IOCEACS)) outputAC(TURNOFF);

                /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
                 * 7 Feb 17:  added if statement below to correct function of
                 * timed night mode.  Timer should ignore detection of LIGHT for
                 * duration of NIGHT mode.  Dark override must be off and if 
                 * ElapsedTime is less than OnTime implies output is on.
                 * 
                 * 17 Mar 17 (v3.1):  added second condition set to evaluation
                 * ((PINPWM == OUTPUT) && IOCEACS).  This prevents DAYTIME 
                 * detection when appliance override is active after a timed
                 * night cycle.
                 * 
                 * 18 Oct 17 (v4.0):  changed (PINPWM == OUTPUT) evaluation to
                 * (statACoutput).
                 *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
                if ((!(RunFlags & CFGFEDARK) && (ElapsedTime < OnTime))\
                        || ((statACoutput) && IOCEACS)) continue;

                /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
                 * During NGTTIME look for onset of day.  When LIGHT is 
                 * detected start timer and set flag FlagCfrmDay.  
                 * WorkingLightState must be LIGHT for entire timing timer cycle 
                 * to change to DAYTIME; any disruption resets.  This minimizes 
                 * the chance of false trigger when ambient light conditions are 
                 * near dark threshold.
                 * 
                 * If timer expires and still LIGHT set DAYTIME.
                 *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
                if (!(FlagCfrmDay))     //Not waiting to confirm LIGHT...
                {
                    if (WorkingLightState == defLight)  //...but if LIGHT:
                    {
                        ResetSecondsCounter();      //Start timer...
                        FlagCfrmDay = TRUE;         //...and set flag to track
                    }
                }
                else                                //Timing to confirm LIGHT
                {
                    if (SecondsCounter > DAYDELAY)  //Timer has expired
                    {
                        FlagCfrmDay = FALSE;        //Reset flag regardless
                        if (WorkingLightState == defLight)
                        {
                            /*++++++++++++++++++++++++++++++++++++++++++++++++++
                             * Daytime confirmed.  Set DAYTIME flag and turn
                             * output off unless appliance override is on.
                             *++++++++++++++++++++++++++++++++++++++++++++++++*/
                            FlagDayOrNight = DAYTIME;
                            if ((statACoutput) && !(IOCEACS)) outputAC(TURNOFF);
                        }
                    } else if (WorkingLightState == defDark) FlagCfrmDay = FALSE;
                }
            }
        }
#endif
    }
} //End function main


/*Function GetLightState.  Used to provide a -stable- method of reading the 
 * current light state.  The ISR handles the acquisition of the light state when
 * it changes and provides a heuristic method of ensuring the light state reported
 * reflects the pin state.  This approach is used to minimize the effects of jitter
 * that is likely to happen during dawn/dusk or transient lighting conditions.
 * 
 * The light state value can be accessed directly in variable CurrentLightState
 * but there is the chance that it does not accurately reflect the pin value if
 * an interrupt to change the value is happening at the same instant.  This 
 * function helps ensure this does not happen by disabling interrupts during the
 * read of the value.  It is understood that the value on the pin could still be 
 * in the process of changing but it will go through the validation process.
 *
 * Returns:  unsigned char value of:
 *      LIGHT
 *      DARK
 *
 * Parameters:  none
 *
*/
unsigned char GetLightState(void) 
{
    di();
    CurrentLightState = LightState;
    ei();
    return(CurrentLightState);
} //End function GetLightState


/*Function ResetCountLightState.  Safely resets light state count:  ensures that 
 * count is in sync with changes to variable LightState.
 *
 * Returns:  none.
 *
 * Parameters:  none
 *
*/
void ResetCountLightState(void)
{
    di();
    CountLightState = 0;
    ei();
    return;
} //End function ResetCountLightState


/*Function BlinkLights.  Cycles the status LED from current setting to opposite and
 * then back.  BlinkSpeed is FAST or SLOW (2x time of FAST).
 * 
 * This function was updated in version 2.1 to control the status LED instead
 * of the output; hence the function name is legacy.
 *
 * Returns:  none.
 *
 * Parameters:  BlinkSpeed =
 *      SLOW
 *      FAST
 *
*/
void BlinkLights(unsigned char BlinkSpeed)
{
    //Toggle lights to opposite state
    ledFLIP();
    __delay_ms(BLINKDELAY);

    //Double the time for slow option
    if (BlinkSpeed == SLOW) __delay_ms(BLINKDELAY);

    //Return lights to original state
    ledFLIP();

    //This delay ensures that back-back calls have some separation
    __delay_ms(BLINKDELAY);
    return;
} //End function BlinkLights


/*Function ResetSecondsCounter.  Safely resets variable SecondsTimer:  avoids 
 * possible contention that might result from reseting value when ISR for 
 * timekeeping executes.
 * 
 * IMPORTANT:  no attempt is made to synchronize this reset with ISR for time
 * keeping.  Hence, routines that use variable SecondsCounter to time an event 
 * must account for accuracy of +/- 2 seconds.
 *
 * Returns:  none.
 *
 * Parameters:  none
 *
*/
void ResetSecondsCounter(void)
{
    di();
    SecondsCounter = 0;
    ei();
    return;
}


/*Function eepromWriteBytes.  Writes data to EEPROM.  This is a wrapper function
 * for the library function eeprom_write.  Unlike it, eepromWriteBytes allows more
 * than one byte to be written and verifies written data.  This function can be
 * used with any PIC with EEPROM that is supported by the library functions
 * eeprom_write and eeprom_read.
 * 
 * Returns:     0 upon success or non-zero if read verify fails.  The non-zero
 * integer represents the offending byte.  The integer is 1-offset.
 *
 * Parameters:  Wptr - unsigned char integer representing eeprom starting address
 *              Rptr - pointer to data to write
 *              NumBytes - number of bytes to write
 *
*/
unsigned char eepromWriteBytes(unsigned char Wptr, unsigned char *Rptr, unsigned char NumBytes)
{
    unsigned char SavBytes;

    //Write data from Rptr to eeprom space pointed to by Wptr.
    //Preserve # bytes written in SavBytes.
    for (SavBytes = NumBytes; (NumBytes); --NumBytes, ++Wptr, ++Rptr) eeprom_write(Wptr, *Rptr);
    while (WR);                         //Wait for last write to complete

    //Verify write; start at end of written block.
    for (--Wptr, --Rptr; (SavBytes); --SavBytes, --Wptr, --Rptr)
    {
        if (eeprom_read(Wptr) != *Rptr) break;
    }

    return(SavBytes);
} //End function eepromWriteBytes


/*Function GetUserAttention.  Blinks lights FAST BLINKASKU times to get user
 * attention
 *
 * Returns:  none.
 *
 * Parameters:  none.
 *
*/
void GetUserAttention()
{
    unsigned char Counter;
    
    for (Counter = BLINKASKU; (Counter); --Counter) BlinkLights(FAST);
    return;
} //End function GetUserAttention


/*Function ConfirmSetting.  Blinks lights to confirm the value of a setting.
 *
 * Returns:  none.
 *
 * Parameters:  integer representing setting value.  No validation checking is 
 * done.
 *
*/
void ConfirmSetting(unsigned char Setting)
{
    for (; (Setting); --Setting) BlinkLights(FAST);

    return;
} //End function ConfirmSetting


/*Function MainMenu.  Allows user to check & set options.
 *
 * Returns:  none.
 *
 * Parameters:  none but expects the following module-level variables to be set.
 *      OnTime, RunFlags, DarkThreshold, OptFlags, SetTemp.
 * These variables can also be modified by this function.
 * 
 *
*/
void MainMenu()
{
    int ambient;
    unsigned char UserChoice,
            CurrentSetting;

    while ((UserChoice = SelectOption(0, OPTSZMMENU, DELAYSELOPT)))
    {
        switch (UserChoice)
        {
            case OPTMDMENU:             //Set run mode
            {
                if (RunFlags & CFGFTMODE) CurrentSetting = OPTTIMED;        //Timed
                else if (RunFlags & CFGRMODEO) CurrentSetting = OPTRESET;   //Reset
                else if (RunFlags & CFGTMODEO) CurrentSetting = OPTTEMPE;   //Temp
                else CurrentSetting = OPTNIGHT;                             //Night

                if ((UserChoice = SelectOption(CurrentSetting, OPTSZDMENU, DELAYSELOPT)))
                {
                    if (UserChoice != CurrentSetting)
                    {
                        RunFlags &= (CFGFNMODE & CFGRMODEF & CFGTMODEF);    //Clear
                        if (UserChoice == OPTTIMED) RunFlags |= CFGFTMODE;
                        else if (UserChoice == OPTRESET) RunFlags |= CFGRMODEO;
                        else if (UserChoice == OPTTEMPE) RunFlags |= CFGTMODEO;
                        RunFlags |= CFGFYCHGD;
                    }
                }
                break;
            }
            case OPTDOMENU:             //Set dark override
            {
                if (RunFlags & CFGFEDARK)
                {
                    CurrentSetting = OPTDKENABL;
                }
                else
                {
                    CurrentSetting = OPTDKDISAB;
                }
                if ((UserChoice = SelectOption(CurrentSetting, OPTSZKMENU, DELAYSELOPT)))
                {
                    if (UserChoice != CurrentSetting)
                    {
                        RunFlags ^= CFGFEDARK;
                        RunFlags |= CFGFYCHGD;
                    }
                }
                break;
            }
            case OPTTMMENU:             //Set on-time
            {
                if ((UserChoice = SelectOption(OnTime, OPTSZHMENU, DELAYSELOPT)))
                {
                    if (UserChoice != OnTime)
                    {
                        OnTime = UserChoice;
                        RunFlags |= CFGFYCHGD;
                    }
                }
                break;
            }
            case OPTORMENU:             //Set appliance override
            {
                if (RunFlags & CFGFEAPPL) CurrentSetting = OPTAOENABL;
                else CurrentSetting = OPTAODISAB;
                if ((UserChoice = SelectOption(CurrentSetting, OPTSZAMENU, DELAYSELOPT)))
                {
                    if (UserChoice != CurrentSetting)
                    {
                        RunFlags ^= CFGFEAPPL;
                        RunFlags |= CFGFYCHGD;
                    }
                }
                break;
            }
            case OPTIVMENU:             //Set invert
            {
                if (RunFlags & CFGFEIVRT)
                {
                    CurrentSetting = OPTIENABL;
                }
                else
                {
                    CurrentSetting = OPTIDISAB;
                }
                if ((UserChoice = SelectOption(CurrentSetting, OPTSZIMENU, DELAYSELOPT)))
                {
                    if (UserChoice != CurrentSetting)
                    {
                        RunFlags ^= CFGFEIVRT;
                        RunFlags |= CFGFYCHGD;
                    }
                }
                break;
            }
            case OPTDCMENU:             //Check dark threshold
            {
                GetUserAttention();
                ResetSecondsCounter();
                DarkThreshold = RtDarkThreshold;
                while (SecondsCounter < DELAYDKCHK)
                {
                    if (GetLightState() == LIGHT)
                    {
                        ledON();
                    }
                    else
                    {
                        ledOFF();
                    }
                    __delay_ms(DKCHKDELAY);
                }
                ledOFF();               //Ensure status LED off
                DarkThreshold = DFLTDKTHD;
                GetUserAttention();
                break;
            }
            case OPTDSMENU:             //Set dark threshold
            {
                GetUserAttention();
                ResetSecondsCounter();
                while (SecondsCounter < DELAYDKSET);
                di();                   //Prevent value change during evaluation & assignment
                if ((SensedLightValue > MINTDKTHD) && (SensedLightValue < MAXTDKTHD))
                {
                    RtDarkThreshold = SensedLightValue;
                    ei();               //Re-enable interrupts
                    RunFlags |= CFGFYCHGD;
                    GetUserAttention(); //Confirms valid
                }
                ei();

                break;
            }
            case OPTSTMENU:             //Set temp
            {
                if (!(UserChoice = SelectOption(SetTemp, OPTSZTMENU, \
                        DELAYSELOPT)))
                {
                    di();
                    ambient = isrTemp;
                    ei();
                    UserChoice = (unsigned char) (ambient - TMPADCZER) \
                            / TMPADCOEF;
                }
                if (UserChoice != SetTemp)
                {
                    SetTemp = UserChoice;
                    RunFlags |= CFGFYCHGD;
                }
                break;
            }
            case OPTTEMENU:             //Set temp evaluation
            {
                if (OptFlags & CFGOFTMPL) CurrentSetting = OPTTEMPL;
                else CurrentSetting = OPTTEMPG;
                if ((UserChoice = SelectOption(CurrentSetting, OPTSZEMENU, \
                        DELAYSELOPT)))
                {
                    if (UserChoice != CurrentSetting)
                    {
                        OptFlags ^= CFGOFTMPL;
                        RunFlags |= CFGFYCHGD;
                    }
                }
                break;
            }
            case OPTRDMENU:             //Restore defaults
            {
                GetUserAttention();
                if (SelectOption(0, OPTSZRMENU, DELAYSELOPT) == OPTRDYES)
                {
                    OnTime = DFLTONTIME;
                    RunFlags = DFLTFLAGS;
                    RtDarkThreshold = DFLTDKTHD;
                    SetTemp = DFLTTEMP;
                    OptFlags = DFLTOPTFLG;
                    RunFlags |= CFGFYCHGD;
                    GetUserAttention();
                }
                break;
            }
            case OPTSVMENU:             //Display software version (v3.1)
            {
                ConfirmSetting(VERMAJOR);
                __delay_ms(SVCHKDELAY);
                BlinkLights(SLOW);
                __delay_ms(SVCHKDELAY);
                ConfirmSetting(VERMINOR);
                __delay_ms(SVCHKDELAY);
                break;
            }
        }
        BlinkLights(SLOW);              //Let user know return to main menu
    }
    return;
} //End function MainMenu


/*Function SelectOption.  Allows user to select a menu item or set an option.
 * 
 * In version 4.1 changed behavior of WaitTime parameter.  In prior versions
 * this value specified how long the user had to complete option selection.
 * Starting with this version it determines how long to wait for the user to make
 * each entry; the timer is reset when input is detected.  This has the effect of
 * giving the user indefinite time to make a selection as long as each gesture
 * happens in WaitTime.  This change was made to ensure that high value selections
 * can be made in difficult lighting conditions.
 * 
 * In version 2.1 removed fast toggle of diags pin (status LED) since now used
 * for menu navigation.
 *
 * Returns:  integer representing valid menu item / option selection or, 0 if 
 * no / invalid option selected.
 *
 * Parameters:
 *      OptValue=current setting to display to user (e.g. light blinks).  If 0
 *          then skip current setting display.
 *      MaxValue=maximum input (e.g. # gestures)
 *      WaitTime=# seconds for user input
 *      
 * 
*/
unsigned char SelectOption(unsigned char OptValue, unsigned char MaxValue, 
        unsigned char WaitTime)
{
    /*Display current setting for user if > 0; pause before & after for readability.
     * 
     */
    if (OptValue)
    {
        __delay_ms(CFRMDELAY);
        ConfirmSetting(OptValue);
    }

    /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     * Allow user to enter setting.  Acquisition loop waits WaitTime seconds and
     * lights blink for each valid gesture.  A valid gesture is represented by
     * two light state changes - e.g. light->dark->light or vice-versa.
     *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    OptValue = 0;
    ResetCountLightState();
    ResetSecondsCounter();
    while (SecondsCounter < WaitTime)
    {
        if ((CountLightState) && !(CountLightState & 0x01))
        {
            ResetCountLightState();
            ++OptValue;
            BlinkLights(FAST);
            ResetSecondsCounter();
        }
    }

    //Zero returned if invalid value set.
    if (OptValue > MaxValue) OptValue = 0;
    return(OptValue);
} //End function SelectOption


/*Function ResetCountLineState.  Safely resets line state count.
 *
 * Returns:  none.
 *
 * Parameters:  none
 *
*/
void ResetCountLineState(void)
{
    di();
    CountLineState = 0;
    ei();
    return;
} //End function ResetCountLineState


#ifdef _16BIT
/*Function uint16_sqrt.  16-bit integer square root.  Credit to Jack W. Crenshaw,
 * http://www.embedded.com/electronics-blogs/programmer-s-toolbox/4219659/Integer-Square-Roots
 *
 * Returns:  unsigned int (16-bit) representing integer square root of Square.
 *
 * Parameters:  uint16_t 16-bit integer value to take square root of.
 * 
 * Requires:  stdint.h.
 *
*/
unsigned int uint16_sqrt(uint16_t Square)
{
    unsigned char Counter;
    uint16_t Root = 0,
             Remainder = 0;

    for (Counter = 0; (Counter < (sizeof(Square)*8/2)); Counter++)
    {
        Root <<= 1;
        Remainder = ((Remainder << 2) + (Square >> ((sizeof(Square)*8)-2)));
        Square <<= 2;
        Root++;
        if (Root <= Remainder)
        {
            Remainder -= Root;
            Root++;
        }
        else
        {
            Root--;
        }
    }
    return ((unsigned int) (Root >> 1));
} //End function uint16_sqrt
#endif


#ifdef _24BIT
/*Function uint24_sqrt.  24-bit integer square root.  Credit to Jack W. Crenshaw,
 * http://www.embedded.com/electronics-blogs/programmer-s-toolbox/4219659/Integer-Square-Roots
 *
 * Returns:  unsigned int (24-bit) representing integer square root of Square.
 *
 * Parameters:  uint24_t 24-bit integer value to take square root of.
 * 
 * Requires:  stdint.h.
 *
*/
uint24_t uint24_sqrt(uint24_t Square)
{
    unsigned char Counter;
    uint24_t Root = 0,
             Remainder = 0;

    for (Counter = 0; (Counter < (sizeof(Square)*8/2)); Counter++)
    {
        Root <<= 1;
        Remainder = ((Remainder << 2) + (Square >> ((sizeof(Square)*8)-2)));
        Square <<= 2;
        Root++;
        if (Root <= Remainder)
        {
            Remainder -= Root;
            Root++;
        }
        else
        {
            Root--;
        }
    }
    return ((uint24_t) (Root >> 1));
} //End function uint24_sqrt
#endif


/*Function CalibrateADC.  Derives an adjustment factor that can be applied to 
 * known constants that ADC values will be evaluated against.  This is necessary
 * when the ADC's positive voltage reference is connected to Vdd and is imprecise.
 * Even when Vdd is supplied by a regulated supply a variance of 100mV from unit 
 * to unit @ 5V can produce a measured delta of 20 at full scale since the 
 * resolution is ~ 5mV.
 * 
 * This function does not require interrupts to be disabled and background
 * peripheral servicing can continue but, it should not be called in timing
 * sensitive sections of the foreground application.  The FVR is used to provide
 * an accurate input value to derive the adjustment factor so it should not be in
 * use by other peripherals running in the background.
 * 
 * The configuration of the ADC and FVR are restored to their calling state when
 * done.  Settling time should be allowed before resuming use.
 *
 * Returns:  non-zero signed integer representing adjustment value to apply to 
 * constants that will be compared to ADC values.  The formula to derive a
 * constant value adjusted for variation in the positive reference is
 *      Ac = C + (C / Af) where:
 *          Af=Adjustment factor
 *          C=Constant
 *          Ac=Adjusted constant
 * 
 * Zero is returned if calibration failed.
 *
 * Parameters:  none.
 * 
 * Requires:  n/a
 *
*/
int CalibrateADC()
{
    unsigned char FVRCONsave,
            ADCON0save,
            ADCON1save,
            ADACTsave;
    int MeasuredDelta,
            AdjFactor,
            AvgValue;
    union {
        unsigned char result[sizeof(int)];
        int           value;
    } Andi;

#ifndef _SIMUL
    //Save current FVR config & setup for ADC calibration
    FVRCONsave = FVRCON;
    FVRCON = CADCFVR;
    __delay_ms(CADCTFVR);               //Wait for stabilization

    //Save current ADC config & setup for ADC calibration
    ADCON0save = ADCON0;
    ADCON1save = ADCON1;
    ADACTsave = ADACT;
    ADCON0 = CADCCON0;
    ADCON1 = CADCCON1;
    ADACT = CADCCON2;

    //Perform AD conversion of FVR to calibrate
    __delay_us(CADCTCON);               //Ensure plenty of time for FVR to settle on initial connection
    for (AdjFactor = CADCCALSPL, AvgValue = 0; (AdjFactor); --AdjFactor)
    {
        __delay_us(CADCTACQ);           //Refresh sample & hold cap
        GO_nDONE = TRUE;                //Start conversion
        while (GO_nDONE);               //Wait for completion
        Andi.result[RESULTHI] = ADRESH;
        Andi.result[RESULTLOW] = ADRESL;
        AvgValue += Andi.value;
    }
    AvgValue /= CADCCALSPL;

    //Calculate adjustment factor
    if (MeasuredDelta = (AvgValue - CADCREFVAL))
    {
        AdjFactor = (AvgValue / MeasuredDelta);
    }
    else
    {
        AdjFactor = CADCREFVAL;
    }

    //Check calibration:  applying adjustment factor should equal ADCREFVAL
    MeasuredDelta = AvgValue - (CADCREFVAL + (CADCREFVAL / AdjFactor));
    //Take absolute value...
    if (MeasuredDelta < 0)
    {
        MeasuredDelta = ((MeasuredDelta ^ 0xFFFF) + 1);
    }
    //...and difference should be less than ADCMAXDIF
    if (MeasuredDelta > CADCMAXDIF) AdjFactor = 0;

    //Restore FVR & ADC peripherals to calling state
    FVRCON = FVRCONsave;
    ADCON0 = ADCON0save;
    ADCON1 = ADCON1save;
    ADACT = ADACTsave;
#else
    AdjFactor = CADCREFVAL;
#endif

    return(AdjFactor);
} //End function CalibrateADC


/*Function outputAC.  Performs initial calibration of ADC & current offset from
 * Hall Effect sensor (both are collectively referred to as calibration) and 
 * controls turn on / off of AC output.  Calibration is also performed
 * prior to turn-on to compensate for drift & temperature.
 *
 * Returns:  SUCCESS if command completes successfully.
 * 
 *           FAIL is returned if calibration fails.  Variable statFAULT will also 
 *           be set which can be cleared by the caller.  This 
 *           function will also clear statFault on a subsequent call if calibration
 *           succeeds and OVER_TEMP is FALSE.
 * 
 *           FAIL is also returned when a TURNON command is issued and OVER_TEMP 
 *           is TRUE.  This cannot be cleared in software; the unit must cool
 *           down.
 * 
 *           IMPORTANT (v4.0):  variable statACoutput will always be set to reflect
 *           the command issued regardless of success.  This is done because
 *           the logic in the timing control loop expects fault conditions to
 *           be handled in the background and relies on the EXPECTED AC output
 *           state for proper control.  Prior versions used the TRIS state of 
 *           the PWM pin to serve this function (the CWG auto-shutdown bit 
 *           indicated fault conditions).
 * 
 *           With the introduction of a 2-stage output switch reliance on HW
 *           status bits is not possible.  Bit flag statACoutput emulates the
 *           prior HW function and avoids a rewrite of the timing control loop.
 *
 * Parameters:  unsigned char Command set to one of the following values:
 *                  TURNON=turn AC output on
 *                  TURNOFF=turn AC output off
 *                  INIT=perform calibration only and take no precautions to
 *                  preserve ADC & FVR state.  Assumes timer4 & timer0 are off.
 *                  This command should only be used during initialization of the
 *                  timer control program.
 * 
 *                  Calling the function with any other value will default to
 *                  INIT.
 * 
 * Requires:  binary.h
 *
*/
unsigned char outputAC(unsigned char Command)
{
    int ACoffsetMin,
            ACoffsetMax;
    unsigned char Counter,
            Retval = FAIL;
    union {
        unsigned char result[sizeof(int)];
        int           value;
    } Adc;                              //Facilitates read of ADC value as integer

    /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     * No checking is done for the OFF command.  The ACS function and flag are
     * both cleared irrespective of being enabled.  Doing so doesn't harm 
     * anything and ensures that if the function is enabled the proper delay
     * & initialization occur.
     * 
     * If PINRELAY is true it's assumed that SS switch is off.  Hence, the SS
     * is turned on to provide a 'soft' off commutation for the relay.
     *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    if (Command == TURNOFF)
    {
        if (PINRELAY)
        {
            pwmON();
            __delay_ms(RLY_DELAY_SOFF);
            relayOFF();
            __delay_ms(RLY_DELAY_EOFF);
        }

        pwmOFF();
        DisableACS();
        FlagCfrmACS = FALSE;
        statACoutput = FALSE;

        return(SUCCESS);
    }

    /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     * Since the TURNON command is called during run-time the ADC must be 
     * released from the ISR to avoid contention.  To do so and ensure a clean 
     * restart the control variable ADacqCtrlFlag is set to OACRELADC which is 
     * monitored by the timer4 & 0 ISR code blocks.  A clean shutdown is 
     * confirmed when TMR4ON bit clears and bit OACRMSSTP of ADacqCtrlFlag is set.
     * A while loop is used to wait for confirmation which can take some time and 
     * is dependent on timer0.  Restart is accomplished by setting the TMR4ON 
     * bit.
     * 
     * NOTE:  if the output is already on then nothing is done and function
     * completes with SUCCESS.  This is done to prevent a sub-second power 
     * disruption to the attached appliance which could be damaging.  It might 
     * also effect the success of calibration.  Bit flag statACoutput is not
     * set in this condition since it's assumed that a prior proper turn-on
     * command was issued that would have set.
     * 
     * If OVER_TEMP is TRUE then TURNON will also fail BUT statACoutput will 
     * still be set since it indicates what the output SHOULD be doing.
     *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    if (Command == TURNON)
    {
        if ((PINPWM == INPUT) && (PINRELAY == FALSE))
        {
            if (OVER_TEMP)
            {
                statACoutput = TRUE;
                return(FAIL);
            }
            ADacqCtrlFlag = OACRELADC;      //Request release of ADC
            while ((TMR4ON) || !(ADacqCtrlFlag & OACRMSSTP));   //Wait...
            ADacqCtrlFlag = OACCLEAR;       //...and clear control flag
        }
        else return(SUCCESS);
    }

    /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     * There's no prep work to do for INIT command since this is only to be 
     * called during MCU initialization prior to timers being enabled.  No 
     * preservation of state is required and it's expected by the caller that at
     * the end of calibration the ADC is ready for current / light sense and is
     * connected to current sense pin.
     * 
     * Use while loop for waterfall processing of calibration pass/fail.  For
     * remainder of function variable Retval is used for return value.
     *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    while (TRUE)
    {
        /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
         * Calibrate ADC.  This makes use of the FVR and when called with the ON
         * command will disconnect FVR from DAC which may cause comparator to 
         * set OVER_TEMP TRUE.  Hence, statFault should be cleared upon 
         * successful completion & prior to return.
         * 
         * Then apply adjustment factor to limit constants.  Note that 
         * ACurrentLimit and AcRelayLimit don't require adjustment since the
         * current sensor's output (mV/A) is ratiometric with Vdd but, these 
         * constants are assigned here for consistency and easy update if 
         * required in the future.
         * 
         * In the -very- unlikely probability that OVER_TEMP goes TRUE after 
         * previous check (evaluation for TURNON command) then it's expected
         * that the ISR will turn the output off within mS of completion here.
         *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
        if (!(ADCAdjFactor = CalibrateADC())) break;
        //Apply adjustment factor to limiting values
        ACurrentLimit = ACCURLIM;
        AcRelayLimit = ACCRELYON;
        ACoffsetMin = ACCOFFMIN + (ACCOFFMIN / ADCAdjFactor);
        ACoffsetMax = ACCOFFMAX + (ACCOFFMAX / ADCAdjFactor);
        assignTemp();


#ifndef _SIMUL
        /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
         * Configure ADC for current & light sense and establish current sense 
         * offset.  Offset not necessary for light since both ADC & sense value 
         * are ratiometric to Vdd.
         * 
         * The average of several samples is taken to establish 0A offset.  The 
         * offset must be within limits or calibration fails.  Obviously the AC 
         * output must be off.
         *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
        ADCON0 = ADCMCON0;                  //Configure ADC & connect to current sense pin
        ADCON1 = ADCMCON1;
        ADACT = ADCMCON2;
        __delay_us(ACOFTCON);               //Time for voltage to settle on initial connection
        for (Counter = ACOFASPL, ACurrentOffset = 0; (Counter); --Counter)
        {
            __delay_us(ACOFTACQ);           //Refresh sample & hold cap
            GO_nDONE = TRUE;                //Start conversion
            while (GO_nDONE);               //Wait for completion
            Adc.result[RESULTHI] = ADRESH;
            Adc.result[RESULTLOW] = ADRESL;
            ACurrentOffset += Adc.value;
        }
        ACurrentOffset /= ACOFASPL;         //Take average
        //Check limits
        if ((ACurrentOffset < ACoffsetMin) || (ACurrentOffset > ACoffsetMax)) break;
#else
        ACurrentOffset = ACCOFFNOM;
#endif

        //If this far calibration has passed so update return code & get out
        Retval = SUCCESS;

        break;
    }

    /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     * No clean-up required for INIT command; the ADC is configured & ready for 
     * current & light sense.  For ON command the timer is always re-enabled and
     * statACoutput set, but the AC output & status bits are set accordingly.
     *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    if (Command == TURNON)
    {
        TMR4ON = ENABLE;
        statACoutput = TRUE;
        if (Retval == SUCCESS)
        {
            pwmON();
            statFault = FALSE;
        }
        else statFault = TRUE;
    }

    return(Retval);
} //End function outputAC


/*Function blinkErrorCode.  Displays error code via status LED by cycling on &
 * off to represent the error number.  At the end of the sequence, or when first
 * called, the LED is held on for a longer period of time to mark the beginning 
 * of the code.
 * 
 * This function is intended to be called when a fatal error is encountered; it 
 * will repeat the error code display until the unit is power-cycled.  It does
 * not return.
 *
 * Returns:  none
 *
 * Parameters:  unsigned char ec represents the error number.
 *
*/
void blinkErrorCode(unsigned char ec)
{
    unsigned char code;

    while (TRUE)
    {
        //Mark start of sequence
        ledON();
        __delay_ms(STS_DSP_STRT);
        ledOFF();
        __delay_ms(STS_DSP_BLNK);

        //Blink error code
        for (code = ec; (code); --code)
        {
            __delay_ms(STS_DSP_BLNK);
            ledON();
            __delay_ms(STS_DSP_BLNK);
            ledOFF();
        }

        //Delay before restarting code display
        __delay_ms(STS_DSP_BLNK * 2);
    }
    
} //End function blinkErrorCode


/*Function rsysTime.  Resets system time.
 * 
 *
 * Returns:  none
 *
 * Parameters:  n/a
 *
*/
void rsysTime()
{
    di();
    ElapsedSeconds = 0;
    ElapsedMinutes = (unsigned int) ElapsedSeconds;
    ei();
    Flag24h = FALSE;

    return;
} //End function rsysTime


/*Function InterruptExec.  MPU ISR (Interrupt Service Routine).  Manages all 
 * enabled interrupts:
 *      -Timer 4, performs AD acquisition for current and light.
 *      -Timer 0, controls sensor data acquisition, AC RMS calculation & CWG 
 *       shutdown, and maintains clock tick timer.
 *      -Timer 1, maintains chronological time via tuning fork oscillator.
 * 
 * Since these timers are collectively & always enabled, and no other interrupts
 * are used, the check for the IE flag is omitted to reduce execution time.
 * 
 * Timer servicing is ordered from highest->lowest execution priority, which is
 * determined based on frequency, execution time, and the likelihood of a 
 * collision (e.g. simultaneous interrupt).
 *
 * Returns:  none
 *
 * Parameters:  none
 *
*/
void interrupt InterruptExec(void)
{
    static unsigned int TempSum = 0;    //Sum of temp samples
#ifdef _24BIT
    static uint24_t ACsum = 0,          //Sum of squared AC current values
            ACsquare;                   //Square of instantaneous AC current value
#endif
#ifdef _16BIT
    static uint24_t ACsum = 0;          //Sum of squared AC current values
#endif
    static bit flagTmSSoff;             //True if switched to relay & timing to turn SS off
    static unsigned char SensedLightState,  //Light state obtained from ADC
            PendedLightState,           //Light state value pending successive sample confirmation
            SensedLineState,            //Line state read from pin
            LineSampleCount = 0,        //Successive sample counter for AC line sense
            LightSampleCount = 0,       //Successive sample counter for light ADC sense
            AcSampleCount = 0,          //Count of # of RMS current samples that have exceeded ACSCYCNT
            RySampleCount = 0,          //Count of # of RMS current samples that have exceeded ACSRYCNT
            AcqStep = ASTPBLANK,        //Bit flags to control current & light AD acquisition steps
            CurrentSampleCount = 0,     //Counter for # of instantaneous current samples taken
            CurrentLineState = NOTBVAL, //Last validated line state
            TempSampleCount = TMPSPLCNT;    //Counter for # of temp samples taken
#ifdef _24BIT
    static unsigned int ACrms;          //True RMS AC current value
#endif
#ifdef _16BIT
    static unsigned int ACsquare,       //Square of instantaneous AC current value
                        ACrms;          //True RMS AC current value
#endif

    static union {
        unsigned char byte[sizeof(int)];
        int           value;
    } ADCResult;                        //Facilitates read of ADC value as integer

    /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     * Timer4 control block:  AD acquisition for current, light and temperature.
     * 
     * Calculating the AC RMS current value is based on taking the square root 
     * of n samples squared & divided by n.  The PIC, being a RISC processor, 
     * incurs substantial overhead to perform these calculations.  Hence, the 
     * math is spread over several steps in both the Timer4 & Timer0 ISR blocks.
     * 
     * Timer4 handles the acquisition, squaring, and summing.  90 samples are 
     * taken for each RMS calculation which equates to a sample every 4 degrees
     * at 60Hz (185.185uS/4 or 46.296uS/degree).  Timer4 is set to run at the 
     * closest value to 1 degree: 46.25uS.  In this way the work of acquiring, 
     * squaring, and summing the current values is broken into four manageable 
     * steps that provide ample time for user and other ISR execution.  This 
     * scheme also accommodates light & temp sampling.
     * 
     * The variable CurrentSampleCount keeps track of the number of current
     * samples taken.  When it reaches zero the Timer0 ISR completes the RMS 
     * calculation & evaluates for over-current control.  No further current 
     * sampling or math occurs until CurrentSampleCount is reset by Timer0.
     * This reduces ISR overhead and ensures adequate processing time for 
     * user land code.
     * 
     * The char AcqStep is bit-mapped and controls the steps.  Each time the ISR 
     * is serviced the high bit is shifted left.  When step 4 is reached it's 
     * reset.  Work done in the steps:
     * 1:  AD conversion on current or temp
     * 2:  for current read & square obtained current value, for temp sum AD
     *     value if sample size hasn't been reached, start light acquisition
     * 3:  AD conversion on light, sum squared current value
     * 4:  read & evaluate obtained light value, start current or temp
     *     acquisition
     * 
     * Temperature was added in version 6.0.  Temp is only sensed when the 
     * output is off (current sensing isn't required).  Given the performance
     * overhead of this block the modifications to support temp were minimized
     * and priority in conditional blocks is given to current sensing.
     * 
     * Object TempSampleCount controls the sampling process and number of
     * samples taken to derive an average.  No conversion to an actual temp is 
     * done and is left for foreground code to handle including any adjustment 
     * factor for ADC calibration.  This value is placed in module object 
     * isrTemp.
     * 
     * All four steps in AcqStep are utilized:
     * 1:   statement added in conditional block evaluating ADacqCtrlFlag to
     *      ensure ADC is connected to current pin when sampling resumes - 
     *      because calibration routines save ADC register states.
     * 2:   conditional block else if (!(statACoutput) && (TempSampleCount))
     *      added after block for CurrentSampleCount to sum temp sample.
     * 3:   Calculate average temp when sample size reached and assign to
     *      module object isrTemp.  FlagTemp is expected to be set TRUE at boot 
     *      and by the timer0 code block handling the ADacqCtrlFlag semaphore so 
     *      that temp conversion actually gets a temp value and not current 
     *      (e.g. first sample is discarded.  This can happen on the first pass
     *      thru the acquisition loop when switching from current to temperature
     *      sampling.
     * 4:   ADC is connected to current sense pin only when statACoutput is
     *      TRUE; otherwise it's connected to the temp sense pin.
     *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    if (TMR4IF)
    {
        AcqStep <<= 1;                  //Goto next step
        /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
         *                      Acquisition Step One
         +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
        if (AcqStep & ASTPSTCUR)
        {
            /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
             * If flag is set stop timer and set control variables for a clean 
             * restart.  This is done to release ADC for calibration or other
             * use.  When done, the user must ensure the ADC is configured 
             * properly for current / light sense in this routine and restart
             * the timer.
             * 
             * By default the ADC is connected to the current pin since 
             * ADacqCtrlFlag indicates an AC output turn-on is about to occur.
             * 
             * ADacqCtrlFlag is acknowledged by the timer0 code block because it
             * has clean-up activity too.
             *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
            if (ADacqCtrlFlag)
            {
                TMR4ON = DISABLE;
                TMR4 = 0;               //Clear timer
                AcqStep = ASTPBLANK;    //Start with step 1 when timer re-enabled
                CurrentSampleCount = 0; //Clear current sample set
                ADCON0 = ADCINPCR;      //Connect ADC to current pin
            }
            else
            {
                GO_nDONE = TRUE;            //Start next conversion
                if (CurrentSampleCount) --CurrentSampleCount;
            }
        }
        /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
         *                      Acquisition Step Two
         +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
        else if (AcqStep & ASTPRDCUR)
        {
            //Connect ADC to light pin & fetch result
            ADCON0 = ADCINPLT;
            ADCResult.byte[RESULTHI] = ADRESH;
            ADCResult.byte[RESULTLOW] = ADRESL;

            //Current calculation
            if (CurrentSampleCount)
            {
                ADCResult.value -= ACurrentOffset;      //Remove offset

                //Take absolute value
                if (ADCResult.value < 0) ADCResult.value = ((ADCResult.value ^ 0xFFFF) + 1);

#ifdef _16BIT
                //Ensure we don't exceed a 16-bit integer when squared
                if (ADCResult.value > ACCURMAX) ADCResult.value = ACCURMAX;
                ACsquare = (ADCResult.value * ADCResult.value);
#endif
#ifdef _24BIT
                ACsquare = (uint24_t) ADCResult.value * (uint24_t) ADCResult.value;
#endif
#ifdef _SIMUL
                NOP();
#endif
            }
            //Get temp only when output off
            else if ((!(statACoutput)) && (TempSampleCount))
            {
                TempSum += (unsigned int) ADCResult.value;
                --TempSampleCount;
            }
        }
        /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
         *                      Acquisition Step Three
         +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
        else if (AcqStep & ASTPSTLGT)
        {
            GO_nDONE = TRUE;            //Start light value conversion
#ifdef _24BIT
            //Make sure ACsum doesn't overflow!
            if ((CurrentSampleCount) && (ACsum < (uint24_t) ACCSQSMMAX))
                ACsum += ACsquare;
#endif
#ifdef _16BIT
            if (CurrentSampleCount) ACsum += ACsquare;
#endif
            else if ((!(statACoutput)) && !(TempSampleCount))
            {
                /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
                 * Temperature sampling when output is off.  Discard first 
                 * sample set when resuming temp sampling.
                 *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
                if (FlagTemp) FlagTemp = FALSE;
                else isrTemp = TempSum / TMPSPLCNT;

                TempSum = 0;
                TempSampleCount = TMPSPLCNT;
            }

#ifdef _SIMUL
            NOP();
            NOP();
#endif
        }
        /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
         *                      Acquisition Step Four
         +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
        else
        {
            if (statACoutput) ADCON0 = ADCINPCR;    //Connect ADC to current pin
            else ADCON0 = ADCINTMP;                 //Connect ADC to temp pin
            
            ADCResult.byte[RESULTHI] = ADRESH;
            ADCResult.byte[RESULTLOW] = ADRESL;

            /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
             *                          Measure light
             * 
             * Evaluate light value and determine light state.  Always assign
             * ADC value to SensedLightValue for reference in main program.
             * 
             *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
            SensedLightValue = ADCResult.value;
            if (ADCResult.value > DarkThreshold)
            {
                SensedLightState = LIGHT;
            }
            else
            {
                SensedLightState = DARK;
            }
            /*
             * Successive sampling technique to eliminate jitter in readings.
             * A change in light state must remain stable for LGTSPLCNT
             * iterations of the ISR before confirming state.  Also store 
             * sensed value from ADC to support setting new dark threshold.
             * 
             */
            if (LightSampleCount)   //Light state changed & now sampling
            {
                ++LightSampleCount;
                if (PendedLightState != SensedLightState)
                {
                    LightSampleCount = 0;
                }
                else
                {
                    if (LightSampleCount == LGTSPLCNT)
                    {
                        if ((LightState = PendedLightState) == LIGHT)
                        {
                            PWM6DCH = LEDDCHIH;
                            PWM6DCL = LEDDCHIL;                            
                        }
                        else
                        {
                            PWM6DCH = LEDDCLOH;
                            PWM6DCL = LEDDCLOL;                            
                        }
                        LightSampleCount = 0;
                        ++CountLightState;
                    }
                }
            }
            else                        //If light state change initiate sampling
            {
                if (SensedLightState != LightState)
                {
                    PendedLightState = SensedLightState;
                    LightSampleCount = 1;
                }
            }
            AcqStep = ASTPBLANK;        //Start acquisition steps over
        }
        TMR4IF = FALSE;
    }

    /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     * Timer0 control block.  Controls data acquisition:
     *  + AC connectivity sense for appliance override feature
     *  + Current AD acquisition cycles
     *  + RMS AC current calculations at end of AD acquisition cycle and CWG 
     *    auto-shutdown
     *  + Maintains LEDTicks used to control LED on/off time in timer run 
     *    loop to indicate various status
     *  + Maintains ACSTicks used as limit timer for detecting user switch
     *    events in appliance override feature
     *  + Temperature AD acquisition cycles
     * 
     * Timer0 runs at 50Hz (20mS).
     * 
     * This block was modified for version 6.0 to support temperature sensing
     * when the output is off.  The value of CurrentSampleCount continues to
     * drive the flow of current sampling & control and now incorporates temp
     * sensing as a lower priority.
     * 
     * The current block is only executed when the output is on (statACoutput
     * is TRUE) and only resets CurrentSampleCount within this block - necessary
     * since timer4 flow control relies on this too.
     * 
     * The ADacqCtrlFlag semaphore block includes additional variable 
     * assignments for temperature so that temp sample resumes appropriately
     * when the output is turned off.  This includes assigning TMPINVALD to
     * isrTemp so that immediately following a turn-off the foreground code
     * can identify a valid temp.
     *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    if(TMR0IF)
    {
        ++LEDTicks;
        if (ACSTicks) --ACSTicks;

        /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
         * Monitor for AC line connectivity state change indicated by IOCFACS.
         * 
         * De-bounce logic confirms state change by requiring new state to be
         * held for LINSPLCNT time.  Also the 'new' state at the end of this time
         * must be opposite the previous stable state or the state change is 
         * ignored.  This guards against situation where state change occurs, 
         * state changes back mid-way thru timing cycle, then remains stable for 
         * entire timing cycle.
         *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
        if (IOCFACS)                    //Line state changed
        {
            SensedLineState = (unsigned char) PINACS;
            LineSampleCount = 1;
            ClearACS();
        }
        else                            //No line state change
        {
            if (LineSampleCount)        //But if previously a change...
            {
                if (++LineSampleCount > LINSPLCNT)  //...and no more state changes for LINSPLCNT time...
                {
                    if (SensedLineState != CurrentLineState)    //...and line state opposite...
                    {
                        ++CountLineState;   //...then confirm & increment.
                        CurrentLineState = SensedLineState;
                    }
                    LineSampleCount = 0;    //Reset for next event
                }
            }
        }

        /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
         * AC RMS current calculations, over-current, and sampling control.
         * When CurrentSampleCount reaches zero a full cycle of samples is
         * ready for evaluation -OR- the output is off.
         *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
        if (!(CurrentSampleCount))
        {
            /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
             * If ADacqCtrlFlag is TRUE set control variables for a clean 
             * restart of both current and temperature sampling.  This control 
             * logic expects that ADacqCtrlFlag is set at the beginning of 
             * output turn-on which means that temp sensing stops and current 
             * sensing resumes.
             * 
             * Because CurrentSampleCount will be reset for a fresh sample set 
             * this code block will not execute again until ADC sampling 
             * resumes.
             *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
            if (ADacqCtrlFlag)
            {
                AcSampleCount = 0;
                RySampleCount = 0;
                ACsum = 0;
                CurrentSampleCount = ACSPLCNT;
                ADacqCtrlFlag |= OACRMSSTP;
                FlagTemp = TRUE;
                TempSampleCount = TMPSPLCNT;
                isrTemp = TMPINVALD;
            }
            else if (statACoutput)
            {
                /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
                 * AC current sampling cycle complete; calculate true RMS, 
                 * evaluate for over-current and relay turn-on, and start next 
                 * sampling cycle (reset CurrentSampleCount).
                 * 
                 * If the RMS value exceeds limit (ACurrentLimit) then 
                 * successive sampling is used to confirm.  Current must exceed 
                 * limit for ACSCYCNT samples before faultOFF() is called to 
                 * turn output off & set fault.
                 *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
#ifdef _SIMUL
                NOP();
#endif
#ifdef _24BIT
                ACrms = (unsigned int) uint24_sqrt(ACsum / ACSPLCNT);
#endif
#ifdef _16BIT
                ACrms = uint16_sqrt((unsigned int) (ACsum / ACSPLCNT));
#endif

                /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
                 * Evaluate for over-current.  When limit reached: switch off.
                 * For protection both SS & relay switches are commanded off
                 * simultaneously (no 'soft' turn-off).  Update status bits &
                 * reset sample counter.
                 *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
                if (AcSampleCount)
                {
                    ++AcSampleCount;
                    if (ACrms < (unsigned int) ACurrentLimit)
                    {
                        AcSampleCount = 0;
                    }
                    else if (AcSampleCount == ACSCYCNT)
                    {
                        faultOFF();
                        AcSampleCount = 0;
                    }
                }
                else
                {
                    if (ACrms > (unsigned int) ACurrentLimit) AcSampleCount = 1;
                }

                /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
                 * Evaluate for relay activation.  Relay must be off and not in
                 * over-current state.
                 * 
                 * When switching to relay an overlap of SS & relay is desired 
                 * for a smooth transition.  RySampleCount is re-purposed as a
                 * countdown timer.
                 * 
                 * It is possible for inconsistent states to arise when an
                 * over-current situation occurs immediately after switching to
                 * relay and before the SS switch is off.  In this event the SS 
                 * switch will be turned off prior to completion of the over-
                 * current evaluation (harmless).
                 * 
                 * It's also possible for a normal turn-off event to occur 
                 * immediately after a switch to relay (while timing for SS
                 * turn-off) that will leave the flag & counter set.  At the 
                 * next turn-on event the flag will be ignored.  The counter will
                 * likely be reset through normal evaluation unless there's an
                 * immediate jump to a current value demanding the relay in 
                 * which case the switch to relay might occur earlier than
                 * normal (harmless).
                 *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
                if (!(PINRELAY) && !(AcSampleCount))
                {
                    if (RySampleCount)
                    {
                        ++RySampleCount;
                        if (ACrms < (unsigned int) AcRelayLimit)
                        {
                            RySampleCount = 0;
                        }
                        else if (RySampleCount == ACSRYCNT)
                        {
                            relayON();
                            flagTmSSoff = TRUE;
                            RySampleCount = SS_DELAY_OFF;
                        }
                    }
                    else
                    {
                        if (ACrms > (unsigned int) AcRelayLimit) \
                            RySampleCount = 1;
                    }
                }
                else if (PINRELAY & flagTmSSoff)
                {
                    if (!(--RySampleCount))
                    {
                        pwmOFF();
                        flagTmSSoff = FALSE;
                    }
                }

                //Sample next AC cycle
                ACsum = 0;
                CurrentSampleCount = ACSPLCNT;
            }
        }

        /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
         * Check for over-temp condition.  Ensure output is off & status set 
         * while over-temp condition exists.  This ensures that output remains 
         * off since turn-on happens in foreground and there's no direct process 
         * coordination.
         * 
         * So this block will execute as long as condition exists but plenty of
         * bandwidth exists.
         *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
        if (OVER_TEMP)
        {
            faultOFF();
        }

        TMR0IF = FALSE;
    }

    /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     * Timer1 control block:  chrono timekeeping.
     * 
     * Provides chronological time keeping services using timer driven by 32.768
     * tuning fork crystal oscillator:
     *  + Maintains elapsed minutes in a 24h cycle and sets Flag24h when a day 
     *    has elapsed.
     *  + Maintains seconds counter (SecondsCounter) for use by other functions 
     *    requiring event timing.  Each increment of 1 represents 2 seconds so
     *    doubles the amount of time that can be tracked in unsigned char.
     *  + Maintains ACSMinutes countdown timer to control on time of appliance
     *    override feature.
     * 
     * 28 Mar 17 (v3.1):  consolidated ElapsedMinutes increment with MININDAY
     * evaluation for efficiency.
     * 
     * Timer1 rolls over at 0.5Hz (every 2S) so seconds counting is done by 2.
     *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    if(TMR1IF)
    {
        ElapsedSeconds += 2;
        ++SecondsCounter;

        if (ElapsedSeconds == SECINMIN)
        {
            ElapsedSeconds = 0;
            if (ACSMinutes) --ACSMinutes;
            if (++ElapsedMinutes == MININDAY)
            {
                ElapsedMinutes = 0;
                Flag24h = TRUE;
            }
        }
        TMR1IF = FALSE;
    }

    return;
} //End function InterruptExec


/*Function Initialize.  Configure MCU peripherals, interrupts, ADC calibration, 
 * and performs startup diagnostics.  Organized into the following major steps:
 *      + Pins
 *      + Timers
 *      + ADC
 *      + CMP/DAC/FVR
 *      + PWM/CWG
 *      + Startup->normal operation
 * 
 * Where possible or practical diagnostics are performed to validate proper 
 * operation of a peripheral or critical function in the unit.  If a fault is
 * detected execution is halted and, the status LED can provide some guidance to
 * the locus of the fault:
 *      + Off=ADC calibration failure
 *      + On=Timer1 did not start (likely failure of crystal oscillator)
 *
 * Returns:  none
 *
 * Parameters:  none
 *
*/
void Initialize(void)
{
    unsigned char Counter;              //Loop counter


    di();                               //No interrupts during initialization

    /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     *                        Initialize bit flags
     *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    statACoutput = FALSE;
    statFault = FALSE;
    FlagCfrmACS = FALSE;
    Flag24h = FALSE;
    FlagTemp = TRUE;


    /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     *                           Configure pins
     * 
     * The status LED's intensity is controlled by PWM6 and changing the duty
     * cycle according to the ambient light sense value (LIGHT or DARK).  To
     * turn the LED on or off the TRIS bit is cleared or set using the macros
     * ledON(), ledOFF() or ledFLIP().  During initialization, before the PWM is 
     * on, the LED pin (PINLED) is set TRUE and controlled as described above.
     *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    //RA2=AC sense
    TRISA2 = INPUT;
    ANSA2 = DIGITAL;
    WPUA2 = DISABLE;
    INLVLA2 = LLTTL;
    //RC0=status LED
    RC0PPS = PPS_M_RC0_NONE;            //Ensure clear after soft resets
    ledON();                            //Diags pin on to indicate power & executing
    ANSC0 = DIGITAL;
    WPUC0 = DISABLE;                    //Weak pull-up disabled when set as input
    ODCC0 = PUSHPULL;
    SLRC0 = MAXSLEW;
    PINLED = TRUE;
    //RC1=AC relay switch drive
    relayOFF();
    TRISC1 = OUTPUT;
    ANSC1 = DIGITAL;
    WPUC1 = DISABLE;                    //Weak pull-up disabled when set as input
    ODCC1 = PUSHPULL;
    SLRC1 = MAXSLEW;
    //RC2=CWG->AC switch drive (PWM5)
    ANSC2 = DIGITAL;
    WPUC2 = DISABLE;                    //Weak pull-up disabled when set as input
    ODCC2 = PUSHPULL;
    SLRC2 = MAXSLEW;
    RC2PPS = PWMPPS;                    //Steer CWGA output to pin
    pwmOFF();                           //No signal on pin until ready
    //RC3=temp sense
    TRISC3 = INPUT;
    ANSC3 = ANALOG;
    //RC4=current sense
    TRISC4 = INPUT;
    ANSC4 = ANALOG;
    //RC5=light sense
    TRISC5 = INPUT;
    ANSC5 = ANALOG;

    __delay_ms(STARTDELAY);             //Wait for Vdd to settle before continuing
    ledOFF();                           //Diags pin off before calibration


    /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     * Make sure that boot was not caused by stack under/over-flow or watchdog
     * reset.  If so halt boot and display error code on status LED.
     * 
     * Otherwise configure watchdog timer and continue boot.
     *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    if ((STKOVF) || (STKUNF)) blinkErrorCode(STS_COD_RSTK);
    if (!((PCON0bits.nRWDT) & (STATUSbits.nTO))) blinkErrorCode(STS_COD_RWDT);

    WDTCON = CFG_WCHD_CON;


    /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     *                          Configure timers
     *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    //Timer0->clk ticks
    T0CON0 = CKTCON0;
    T0CON1 = CKTCON1;
    TMR0H = CKTFREQ;
    TMR0IE = ENABLE;
    //Timer1->chrono time keeping
    T1CON = CHRTCON;
    T1GCON = CHRGCON;
    TMR1IE = ENABLE;
    /*Timer1 count registers are initialized to a value that provides ~ 31mS of
     * delay before timer roles over, calls ISR, and increments seconds counter.
     * Per documentation, this delay is long enough to consider the oscillator 
     * stable.  This delay is relied upon at end of initialization routine as a
     * diagnostic to confirm time keeping.
     */
    TMR1L = CHRINTL;
    TMR1H = CHRINTH;
    //Timer2->PWM5 (AC switch)
    T2CON = PWMTCON;
    PR2 = PWMFREQ;
    TMR2IE = DISABLE;
    //Timer4->AD acquisition sequencing
    T4CON = ADCTCON;
    PR4 = ADCFREQ;
    TMR4IE = ENABLE;
    //Timer6->PWM6 (status LED)
    T6CON = LEDTCON;
    PR6 = LEDFREQ;
    TMR6IE = DISABLE;

    /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     * Perform ADC & current sense calibration.  On fail halt MCU for safety.
     * Otherwise update LED to indicate progress with initialization.
     *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    if (outputAC(INIT) == FAIL) blinkErrorCode(STS_COD_ADCF);
    ledON();

    /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     * Configure Comparator1 to monitor temperature sensor and set CWG auto
     * shutdown.  Requires dedicated use of FVR & DAC to provide accurate voltage
     * representing desired shutdown temperature.
     *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    FVRCON = CMPFVRCON;
    __delay_ms(CMPSTIME);
    DACCON0 = CMPDACC0;
    DACCON1 = CMPDACC1;
    CM1CON0 = CMPCON0;
    CM1CON1 = CMPCON1;

    /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     * Configure PWM/CWG to drive AC switch and PWM for status LED dimming.  No 
     * output until associated timers start.
     * 
     * First assign timers:  timer2->PWM5, timer6->PWM6.
     *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    PWMTMRS = PWMTIMERS;
    //PWM5
    PWM5CON = PWMCON;                   //Config
    PWM5DCH = ((PWMDC & 0x3FC) >> 2);   //Duty cycle - one time assignment
    PWM5DCL = ((PWMDC & 0x003) << 6);
    //PWM6
    PWM6CON = LEDCON;                   //Config
    PWM6DCH = LEDDCHIH;                 //Duty cycle
    PWM6DCL = LEDDCHIL;
    //Configure CWG1 to support PWM5 for auto-shutdown support
    CWG1AS0 = PWMCWGAS0;                //Place CWG in auto-shutdown before completing config
    CWG1AS1 = PWMCWGAS1;
    CWG1CON0 = PWMCWGCON;               //Enabled but auto-shutdown keeps output off
    CWG1CON1 = PWMCWGPOL;
    CWG1CLKCON = PWMCWGCLK;
    CWG1DAT = PWMCWGDAT;
    CWG1STR = PWMCWGSTR;

    /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     *                           Start Timers
     * 
     * Enable interrupts and start time keeping.  Start interrupt enabled timers 
     * first, slowest to fastest, as these drive timing functions & coordinate 
     * underlying system operation.
     * 
     * Before starting [timers for] PWMs validate that timer1 (driven by crystal
     * oscillator) is running.  Wait SETUPDELAY to leave status LED on long 
     * enough for user to tell initialization is progressing while light sensor 
     * and AC continuity sense settle for stable acquisition.  This delay provides 
     * time for crystal oscillator driven timer1 (chrono) to start.
     * 
     * To ensure that chrono timekeeping is available a validation loop is used
     * to confirm that timer1 ISR is incrementing SecondsCounter.  Wait for 
     * CHRONDELAY CHRONINCRM increments and if timer has not advanced halt MCU 
     * since unit will not function.
     *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    PEIE = ENABLE;                      //Enable peripheral interrupts
    ei();                               //Clear global interrupt mask
    TMR1ON = ENABLE;                    //Start chrono timer
    T0EN = ENABLE;                      //Start clk tick timer
    TMR4ON = ENABLE;                    //Start AD acquisition timer
#ifndef _SIMUL
    __delay_ms(SETUPDELAY);
#else
    NOP();
#endif

    //Timer1 validation
    for (Counter = CHRONDELAY; (Counter); --Counter)
    {
        if (SecondsCounter) break;
        __delay_ms(CHRONINCRM);
    }
    if (!(Counter)) blinkErrorCode(STS_COD_RTCF);   //Disable unit if timed out

    /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     *                           Start PWMs
     * 
     * Time keeping validated.  Start timers driving PWMs and enable CWG.  Assign 
     * PWM6 to LED pin to enable dimming.
     *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    TMR2ON = ENABLE;                    //Start PWM timer
    TMR6ON = ENABLE;                    //Start LED timer
    PWMFLAGSD = FALSE;                  //Start CWG
    RC0PPS = LEDPPS;                    //PWM6 drives LED pin

    /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     *                       Finish Initialization
     * 
     * Turn off status LED to indicate initialization is complete.  Add a delay
     * so that subsequent use of status LED is distinct.
     * 
     * Last, force AC continuity sense by setting IOC bit.  This ensures that 
     * system starts with a validated line state.  To ensure this works the 
     * variable CurrentLineState, in timer0 ISR, must be set to a non-binary
     * value.
     *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    ledOFF();                           //Passed diags so turn off
#ifndef _SIMUL
    __delay_ms(CFRMDELAY);
#endif
    SetACS();                           //Force AC continuity sense

    return;
} //End function Initialize



//End of file