/*

DESCRIPTION:
    Main include file for light timer appliance.

SYNTAX:  n/a

KNOWN BUGS:  n/a.

COPYRIGHT:  (c) 2016-2018, Brian Cornell, all rights reserved.

DEPENDENCIES:  none

VERSION:  6.0

MODIFICATION HISTORY:

    13 May 19, BWC, version 6 branch started
    30 Apr 18, BWC, moved config register pragmas to config.h
        + Added function definition blinkErrorCode
        + Added constants to support WDT
        + Added boot fail status codes & LED blink timing constants
        + Formalized confitional compile section and added definitions for 10A &
          20A current sensors and on/off cycle test routine.
    20 Jan 18, BWC, version 5.0 branch started:  see main.c
    1 Jan 18, BWC, updated to use 20A current sensor 100mV/A and to allow 15A
        continuous operation.
    19 Apr 17, BWC, updated 3.x development branch based on release 3.1.
    29 Mar 17, BWC, Released as version 3.1.
    8 Feb 17, BWC, completed version 3.0 and finalized.  Supports hwr170107.
    11 Dec 16, BWC, development version 3.0d started
    1 Dec 16, BWC, development version 2.2d started
    19 Nov 16, BWC, development version 2.1d started
    16 Nov 16, BWC, post 2.0 development release 2.0d (I know, bad versioning) started
    5 Nov 16, BWC, release 2.0 finalized
    6 Sep 16, BWC, release 2.0 started.
    1 May 16, BWC, release 1.4 development started
    17 Apr 16, BWC, first release incorporating flash storage of user settings
    4 Apr 16, BWC, first release using light sense method.
    4 Jan 16, BWC, development release.

*/


/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 * Conditional compilation control:  uncomment to enable / disable feature.
*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
//#define _SIMUL                          //Running in simulator
//#define _CS_10A                         //Using 10A current sensor
#define _CS_20A                         //Using 20A current sensor
//#define _TEST_ON_OFF                    //Boots & enters on-off test cycle


//Software version
#define VERMAJOR    6
#define VERMINOR    0
#define VERBUILD    67


/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 * System wide constants & definitions.  Convention:  S_<TYPE>_<INFO>
 * New convention adopted with version 6.0
 * Types:
 *  + DC=Duty Cycle
 *  + DF=Default value for a configurable parameter
 *  + EN=ENable bit
 *  + RT=RaTio or scaling factor
 *  + LM=LiMiting operational value
 *  + TM=TiMing
 *  + MP=Peripheral port bit MaPping
 *  + CH=character
 *  + IN=INitializing value
 *  + VL=VoLtage
 *  + QY=quantity (count) of something
 *  + CD=command data
 *  + BL=boolean value
 *  + MK=bit MasK
 *  + SZ=size of an object or data type
 * 
 * Info is user defined and should be descriptive.
 *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
#define S_SZ_BYTE           8               //Number of bits in a byte
#define S_MK_U16_L          0x00FF          //Mask to access low byte of 16-bit unsigned
#define S_MK_U16_H          0xFF00          //Mask to access high byte of 16-bit unsigned


/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 *                           PPS & direct pin mapping
 * 
 * For PPS output pin assignments both aliases for pin registers and constants 
 * mapping what peripheral connects to the pin.
 *      Pin register alias convention:  PPS_P_XXX_YY where
 *          XXX=peripheral name
 *          YY=channel or other detail; can be omitted
 *      Peripheral mapping convention:  PPS_M_XXX_YY same as alias
 * 
 * Input pins only require mapping constant.
 * 
 * For pins accessed directly or assigned in a peripheral's register (like ADC)
 * use the following convention:  PIN_X_YYYYYY where
 *          X=Input or Output
 *          YYYYYY=user assigned mnemonic describing function
 *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
#define PPS_M_RC0_NONE      0x00


/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
*                      Binary constants & definitions
*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
#define DARK        0
#define LIGHT       1
#define SLOW        0
#define FAST        1
#define MODENGT     0
#define MODETIM     1
#define DAYTIME     1
#define NGTTIME     0
#define ACLOPEN     0                   //No AC line continuity
#define ACLCLSD     1                   //AC line continuity


/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 * AC current measurement values.  Current constants represent expected ADC value 
 * with Vdd=5V.
*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
#ifdef _CS_20A
#define ACCURLIM    318                 //Maximum allowable AC current ~ 15.5A RMS
#define ACCRELYON   41                  //Current sense value to switch from SS->relay (~2A RMS)
#endif

#ifdef _CS_10A
#define ACCURLIM    400                 //Maximum allowable AC current ~ 10.6A RMS
#define ACCRELYON   80                  //Current sense value to switch from SS->relay (~2A RMS)
#endif

#define ACCSQSMMAX  16500000            //Maximum sum of squares value to prevent overflow
#define ACCOFFMIN   409                 //Minimum allowable current sense 0A output voltage (2.0V)
#define ACCOFFNOM   512                 //Nominal current sense 0A output voltage (2.5V)
#define ACCOFFMAX   553                 //Maximum allowable current sense 0A output voltage (2.7V)
#define ACOFTCON    80                  //# uS to wait after connecting ADC to current sense pin
#define ACOFASPL    5                   //# samples to use for current sense offset
#define ACOFTACQ    6                   //# uS for acquisition
#define ACSPLCNT    90                  //# current samples per cycle
#define ACSCYCNT    120                 //# sampling cycles ACCURLIM can be exceeded before turning off output
#define ACSRYCNT    90                  //# sampling cycles ACCRELYON can be exceeded before switching from SS->relay


/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 *                      Light sense control constants
*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
#define LGTSPLCNT   11          //Number of light samples required to confirm value
#define MINTDKTHD   0x000A      //Minimum allowable dark threshold light sense value
#define MAXTDKTHD   0x0399      //Maximum allowable dark threshold light sense value


/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 *                      Temperature measurement
*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
#define TMPSPLCNT   6           //Number of successive temp samples to average
#define TMPINVALD   MAXADCVAL   //Invalid temp
#define TMPADCOEF   2           //For ADC Vref=5, ADC value per C (~ 10mV)
#define TMPADCZER   102         //For ADC Vref=5, ADC value for 0C (~500mV)

//Macro to calculate ADC value of set temp and apply calibration
#define assignTemp() adcSetTemp = TMPADCZER + (TMPADCOEF * (int) SetTemp); \
adcSetTemp = adcSetTemp + (adcSetTemp / ADCAdjFactor)


/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 *      Appliance override (AC line continuity sense) control constants
*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
#define PINACS      RA2                 //AC line continuity sense digital input
#define IOCFACS     (IOCAF & 0x04)      //Flag to indicate ACS logic state changed
#define IOCEACS     (IOCAP & 0x04)      //Flag to indicate detection of ACS logic state changes enabled
#define ClearACS()  IOCAF &= 0xFB       //Clears pin interrupt on change flag
#define SetACS()    IOCAF |= 0x04       //Sets pin interrupt on change flag
#define EnableACS() IOCAP |= 0x04; IOCAN |= 0x04    //Enables IOC +/- edge for PINACS
#define DisableACS() IOCAP &= 0xFB; IOCAN &= 0xFB   //Disables IOC +/- edge for PINACS
#define LINSPLCNT   10                  //LineSampleCount must be > to acknowledge line state change
#define ACSONDELAY  250                 //Delay, in 20mS increments, for enablement of ACS (max=255)
#define ACSSWDELAY  250                 //Max time, in 20mS increments, between switch events to turn ACS on
#define ACSONTIME   60                  //Number of minutes appliance override is on


/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
*             AC switch drive PWM / CWG, timer and related settings
*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
//Assign timers to PWMs
#define PWMTIMERS       0x0D                //timer2->PWM5, timer6->PWM6
//AC switch drive config.  Includes timer2, PWM5, CWG1
#define PINPWM          TRISC2              //PWM output control for selected pin
#define PWMPPS          0x08                //CWG (PWM) on RC2
#define pwmON()         PINPWM = OUTPUT     //PWM signal on pin on
#define pwmOFF()        PINPWM = INPUT      //PWM signal on pin off
#define SS_DELAY_OFF    25                  //# of 20mS cycles to wait before SS off after switch to relay

//AC relay switch drive config
#define PINRELAY        RC1                 //Relay on/off control on selected pin
#define relayON()       PINRELAY = TRUE     //PWM signal on pin on
#define relayOFF()      PINRELAY = FALSE    //PWM signal on pin off
#define RLY_DELAY_SOFF  250                 //#mS to wait after SS on before relay off
#define RLY_DELAY_EOFF  250                 //#mS to wait after relay off before SS off

//Macro to turn output off in fault condition & set status
#define faultOFF()      pwmOFF(); relayOFF(); statFault = TRUE

//Timer2->PWM5
#define PWMTCON     0x00        //Off, 1:1 pre-scale, 1:1 post-scale
#define PWMFREQ     79          //Frequency=100kHz
//PWM5 settings
#define PWMCON      0x80        //Enabled, output active high
#define PWMDC       32          //10% (1uS) duty cycle
//CWG1 settings
#define PWMCWGCON   0x80        //Enabled, async steering mode
#define PWMCWGPOL   0x00        //Normal polarity on all channels
#define PWMCWGCLK   0x00        //Clock source is Fosc
#define PWMCWGDAT   0x07        //PWM5 is input
#define PWMCWGSTR   0x01        //CWGxA output has the CWGx data input waveform with polarity control from POLA bit
#define PWMCWGAS0   0xA8        //Auto-shutdown, restart disabled, all pins 0
#define PWMCWGAS1   0x00        //No auto-shutdown sources
#define PWMFLAGSD   CWG1AS0bits.SHUTDOWN    //Indicates/controls PWM auto-shutdown event

//Function outputAC constants
#define OACRELADC   0x01        //Timer4 release ADC (stop) & stop RMS calculations in timer0
#define OACRMSSTP   0x80        //Timer0 block has stopped RMS calculations & reset
#define OACCLEAR    0x00        //Clear control flag
#define TURNOFF     0           //Turn AC output off
#define TURNON      1           //Calibrate & turn AC output on
#define INIT        2           //Calibrate & config ADC for current sense

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
*              Status LED PWM, timer, and related settings
*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
//Status LED drive config.  Includes timer6, PWM6
#define PINLED      RC0         //Diagnostics & testing
#define LEDPPS      0x03        //PWM6 (LED) on RC0
#define ledON()     TRISC0 = OUTPUT //Status LED on
#define ledOFF()    TRISC0 = INPUT  //Status LED off
#define ledFLIP()   TRISC0 ^= TRUE  //Flip LED to opposite state
//Status LED blink rates, in clock ticks, for various conditions
#define DIAGFAULT   15          //Fault
#define DIAGACS     50          //Appliance override active
#define DIAGRUN     100         //Timer running
#define DIAGIDLE    5           //Timer idle
//Timer6->PWM6
#define LEDTCON     0x00        //Off, 1:1 pre-scale, 1:1 post-scale
#define LEDFREQ     159         //Frequency=50kHz (20uS)
//PWM6 settings
#define LEDCON      0x80        //Enabled, output active high
#define LEDDCHI     0x140       //50% DC - use for DAY
//Hard code hi/low bytes for DC to reduce execution time in ISR
#define LEDDCHIH    0x50        //8msb for LEDDCHI
#define LEDDCHIL    0x00        //2lsb for LEDDCHI
#define LEDDCMD     0x040       //10% DC - medium
#define LEDDCLS     0x020       //5% DC - alternate slightly brighter NIGHT
#define LEDDCLO     0x013       //2.97% DC - use for NIGHT
//Hard code hi/low bytes for DC to reduce execution time in ISR
#define LEDDCLOH    0x04        //8msb for LEDDCLO
#define LEDDCLOL    0xC0        //2lsb for LEDDCLO

//AD acquisition configuration & control constants
#define RESULTHI    1           //ADRESH
#define RESULTLOW   0           //ADRESL
#define MAXADCVAL   0x03FF      //Maximum 10-bit value ADC can produce
#define ADCMCON0    0x51        //ANC4 (current) input, ADC enabled
#define ADCMCON1    0xA0        //Right justified, Fosc/32, Vref-=Vss, Vrpos=Vdd
#define ADCMCON2    0x00        //No auto-conversion
#define ADCINTMP    0x4D        //ADC connected to temp pin (RC3)
#define ADCINPCR    0x51        //ADC connected to current pin (RC4)
#define ADCINPLT    0x55        //ADC connected to light pin (RC5)
//Timer4 config for AC acquisition sequencing
#define ADCTCON     0x08        //Off, 1:1 pre-scale, 1:2 post-scale
#define ADCFREQ     184         //Period match value for 46.5uS period
//Timer4 ISR flow control constants
#define ASTPBLANK   0x01        //No step:  ready to start
#define ASTPSTCUR   0x02        //Step 1:  start current conversion
#define ASTPRDCUR   0x04        //Step 2:  read current value
#define ASTPSTLGT   0x08        //Step 3:  start light conversion
#define ASTPRDLGT   0x10        //Step 4:  read light value

//CalibrateADC constants
#define CADCFVR     0x82        //Enable module, disable temp sense, 2.048V->ADC
#define CADCTFVR    60          //# mS to let FVR stabilize
#define CADCCON0    0xFD        //FVR input, ADC module enabled
#define CADCCON1    0xA0        //Right justified, Fosc/32, Vref-=Vss, Vrpos=Vdd
#define CADCCON2    0x00        //No auto-conversion
#define CADCTCON    80          //# uS to wait after connecting ADC to FVR
#define CADCTACQ    6           //# uS for acquisition
#define CADCREFVAL  419         //Epected ADC value in calibration
#define CADCCALSPL  5           //# FVR samples to use for ADC calibration
#define CADCMAXDIF  3           //Maximum calibration difference allowed

//Timer0->clk ticks configuration
#define CKTCON0     0x04        //Disabled, 8-bit, 1:5 post-scale
#define CKTCON1     0x47        //Fosc/4, sync, 1:128 pre-scale
#define CKTFREQ     249         //Frequency=50Hz (20mS)

//Timer1 - Chrono time keeping constants
#ifndef _SIMUL
#define CHRTCON     0b10001000  //Crystal on SOSCO, 1:1 pre-scale, sync w/Fosc/4, stopped
#define CHRINTH     0xFC        //Initial value for TMR1H register
#else
#define CHRTCON     0x30        //Fosc/4, 1:8 pre-scale, stopped
#define CHRINTH     0x00        //Initial value for TMR1H register
#endif
#define CHRINTL     0x00        //Initial value for TMR1L register
#define CHRGCON     0x00        //Gating disabled
#define SECINMIN    60          //Number of seconds in a minute
#define MININHOUR   60          //Number of minutes in a hour
#define MININDAY    1440        //Number of minutes in a day

//Comparator1 configuration and constants for over-temp monitoring
#define CMPFVRCON   0x88        //Enabled, Comparator FVR Buffer Gain is 2x->DAC
#define CMPSTIME    60          //# mS to allow FVR output to stabilize
#define CMPDACC0    0x88        //Enabled, 2.048V Vref from FVR, negative source=Vss
#define CMPDACC1    0x17        //1.472V output for ~ 100C temperature limit
#define CMPCON0     0x96        //CM1 enabled, inverted, hysteresis on, asynchronous
#define CMPCON1     0x2B        //No interrupts, Vdac->C1VP, RC3->C1IN3-
#define OVER_TEMP   C1OUT       //Comparator output:  TRUE if over-temp, otherwise FALSE 

//General timing constants
#define STARTDELAY  500         //# mS to delay initialization
#define SETUPDELAY  2000        //# mS to delay setup @ start
#define CHRONINCRM  100         //# mS in CHRONDELAY increment
#define CHRONDELAY  70          //# of CHRONINCRM increments to wait for chrono to increment
#define BLINKDELAY  600         //# mS to toggle lights to confirm a setting
#define SETCFDELAY  5000        //# mS to wait for user to confirm change to stored settings
#define PRGDELAY    5000        //# mS to wait before entering programming mode
#define CFRMDELAY   1000        //# mS to pause before & after confirming setting
#define DKCHKDELAY  200         //# mS to pause after setting lights during dark threshold check
#define SVCHKDELAY  500         //# mS to pause between & after diaplay of version components
#define NGTDELAY    180         //Wait 360 S before confirming NGTTIME
#define DAYDELAY    180         //Wait 360 S before confirming DAYTIME
#define HOURSDSKDN  48          //Stay on from dusk to dawn
#define HOURSDEFLT  5           //Default on time in hours
#define HOURSMAX    720         //Maximum number of on hours (12)
#define INHIBITNGT  (MININDAY - 30) //# minutes to inhibit flip to NGTTIME from start of prior cycle
#define TMCNTGEST   20          //#mS for fault reset gesture:  TMCNTGEST * DIAGFAULT * 0.02
#define RESETTIME   5           //# minutes output should be OFF in reset mode
#define TEMPTIME    20          //# minutes output should be off between temp mode cycles


//Menu timing & selection constants
#define BLINKASKU   3           //Number of blinks to get user attention
#define DELAYSELOPT 2           //# 2x seconds to wait for option selection
#define DELAYDKCHK  5           //# 2x seconds to let user check dark threshold
#define DELAYDKSET  2           //# 2x seconds to let user set dark threshold

//Main menu options
#define OPTMDMENU   1           //Set run mode
#define OPTDOMENU   2           //Set dark override
#define OPTTMMENU   3           //Set on time
#define OPTORMENU   4           //Set appliance override
#define OPTIVMENU   5           //Set invert
#define OPTDCMENU   6           //Check dark threshold
#define OPTDSMENU   7           //Set dark threshold
#define OPTSTMENU   8           //Set temp
#define OPTTEMENU   9           //Temp > or <
#define OPTRDMENU   10          //Restore defaults
#define OPTSVMENU   11          //Display software version

//Menu option size
#define OPTSZMMENU  11          //Number of options for main menu
#define OPTSZDMENU  4           //Number of options for run mode
#define OPTSZKMENU  2           //Number of options to dark override
#define OPTSZHMENU  12          //Maximum on time in hours
#define OPTSZAMENU  2           //Number of options to set appliance override
#define OPTSZIMENU  2           //Number of options to set invert
#define OPTSZTMENU  40          //Number of options (e.g. max temp) for set temp
#define OPTSZEMENU  2           //Number of options for set temp evaluation
#define OPTSZRMENU  2           //Number of options to restore defaults

//Menu option selection values
//Run mode
#define OPTTIMED    1           //Timed run mode
#define OPTNIGHT    2           //Night run mode
#define OPTRESET    3           //Reset run mode
#define OPTTEMPE    4           //Temperature run mode

//Dark override
#define OPTDKDISAB  1           //Dark override disabled
#define OPTDKENABL  2           //Dark override enabled

//Appliance override
#define OPTAODISAB  1           //Appliance override disabled
#define OPTAOENABL  2           //Appliance override enabled

//Invert
#define OPTIDISAB   1           //Invert disabled
#define OPTIENABL   2           //Invert enabled

//Temp evaluation:  > or <
#define OPTTEMPL    1           //Less than
#define OPTTEMPG    2           //Greater than

//Restore defaults
#define OPTRDYES    2           //Yes restore defaults
#define OPTRDNO     1           //No restore defaults

//Not really a menu option
#define OPTRSTFLT   6           //Number of light state counts that equate to three gestures


/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 *              Constants for storage & retrieval of user settings
 *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
//Mapping for FlashBlock configuration array
#define ROWSIZE     16

//Union offset addresses for configuration settings
#define CFGCAFLAG   0           //Character array offset for object RunFlags
#define CFGCATIME   1           //Character array offset for config on-time
#define CFGCADARK   2           //Character array offset for config dark threshold
#define CFGIADARK   1           //Integer array offset for config dark threshold
#define CFGCATEMP   4           //Character array offset for set temp
#define CFGCAOFLG   5           //Character array offset for object OptFlags

//Bit-mapped flags for object RunFlags:
#define CFGFYCHGD   0x80        //Bit 7=1 config changed
#define CFGFNCHGD   0x7F        //Bit 7=0 config unchanged
#define CFGFTMODE   0x01        //Bit 0=1 timed mode
#define CFGFNMODE   0xFE        //Bit 0=0 night mode
#define CFGFEAPPL   0x02        //Bit 1=1 enable appliance override
#define CFGFDAPPL   0xFD        //Bit 1=0 disable appliance override
#define CFGFEDARK   0x04        //Bit 2=1 enable dark override
#define CFGFDDARK   0xFB        //Bit 2=0 disable dark override
#define CFGFEIVRT   0x08        //Bit 3=1 enable night mode invert
#define CFGFDIVRT   0xF7        //Bit 3=0 disable night mode invert
#define CFGRMODEO   0x10        //Bit 4=1 reset mode ON
#define CFGRMODEF   0xEF        //Bit 4=0 reset mode OFF
#define CFGTMODEO   0x20        //Bit 5=1 temp mode ON
#define CFGTMODEF   0xDF        //Bit 5=0 temp mode OFF

//Bit-mapped flags for object OptFlags:
#define CFGOFTMPL   0x01        //Bit 0=1 temp less than
#define CFGOFTMPG   0xFE        //Bit 0=0 temp greater than

//Default settings
#define DFLTFLAGS   CFGTMODEO   //Default run flags
#define DFLTONTIME  5           //Default on time in hours
#define DFLTDKTHD   0x01EB      //Default dark threshold light sense value
#define DFLTDKTHDH  (DFLTDKTHD >> S_SZ_BYTE)    //Default dark threshold light sense value high byte
#define DFLTDKTHDL  (DFLTDKTHD & S_MK_U16_L)    //Default dark threshold light sense value low byte
#define DFLTTEMP    20          //Default set temp
#define DFLTOPTFLG  CFGOFTMPL   //Default option flags


/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 * System clock & watchdog configuration.
 *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
#define _XTAL_FREQ          32000000    //For __delayxxx macros; must match Fosc
#define CFG_WCHD_CON        0x16        //2s interval, disabled


/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 * Boot failure codes & display timing.
 *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
#define STS_COD_RWDT        1           //Reset due to watchdog
#define STS_COD_RSTK        2           //Reset due to stack over/under flow
#define STS_COD_ADCF        3           //ADC current sense calibration fail
#define STS_COD_RTCF        4           //Real time clock failed to start

#define STS_DSP_STRT        2000        //#mS LED on @ start of code display
#define STS_DSP_BLNK        650         //#mS for blink speed


/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 * Function declarations.
 *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
void interrupt InterruptExec(void);
void Initialize(void);
void main(void);
unsigned char GetLightState(void);
void ResetCountLightState(void);
void BlinkLights(unsigned char);
void ResetSecondsCounter(void);
unsigned char eepromWriteBytes(unsigned char, unsigned char *, unsigned char);
void GetUserAttention(void);
void ConfirmSetting(unsigned char);
void MainMenu(void);
unsigned char SelectOption(unsigned char, unsigned char, unsigned char);
void ResetCountLineState(void);
unsigned int uint16_sqrt(uint16_t);
uint24_t uint24_sqrt(uint24_t);
int CalibrateADC(void);
unsigned char outputAC(unsigned char);
void blinkErrorCode(unsigned char);
void rsysTime(void);



//End of file